#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Metaclass for objects that handle installing packages inside test environments."""

from abc import abstractmethod

from .injectable import Injectable


class BasePackageError(Exception):
    """Base error for package handlers."""


class BasePackage(Injectable):
    """Metaclass for package handlers.

    Packages define tooling stubs needed to install packages inside test environments.
    """

    @abstractmethod
    def _run(self) -> None:
        """Run installer for package."""

    @abstractmethod
    def _setup(self) -> None:
        """Perform setup needed inside test environment to run installer."""



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Configure and control the LXD test environment provider."""

import copy
from typing import Any, Dict

from cleantest.meta._base_configurer import BaseConfigurer, BaseConfigurerError

from .lxd_config import ClientConfig, InstanceConfig, _DefaultSources


class BadClientConfigurationError(BaseConfigurerError):
    """Raised if given client configuration is bad."""


class DuplicateLXDInstanceConfigError(BaseConfigurerError):
    """Raised if two or more instance configurations share the same name."""


class LXDInstanceConfigNotFoundError(BaseConfigurerError):
    """Raised when LXD instance configuration is not found in the registry."""


class LXDConfigurer(BaseConfigurer):
    """Configurer for LXD test environment provider."""

    __configs = {
        InstanceConfig(name=name.replace("_", "-").lower(), source=source)
        for name, source in _DefaultSources.items()
    }
    __client_config = ClientConfig()

    def __new__(cls) -> "LXDConfigurer":
        """Create new LXDConfigurer instance.

        Returns:
            (LXDConfigurer): New LXDConfigurer instance.
        """
        if not hasattr(cls, f"_{cls.__name__}__instance"):
            cls.__instance = super(LXDConfigurer, cls).__new__(cls)
        return cls.__instance

    @property
    def client_config(self) -> ClientConfig:
        """Get LXD client configuration information.

        Returns:
            (ClientConfig): Current LXD client configuration.
        """
        return self.__client_config

    @client_config.setter
    def client_config(self, new_conf: Dict[str, Any]) -> None:
        """Set LXD client configuration.

        Raises:
            BadClientConfigurationError:
                Raised if new configuration contains invalid configuration option.
        """
        valid_config_options = self.__client_config.keys()
        for k in new_conf.keys():
            if k not in valid_config_options:
                raise BadClientConfigurationError(
                    (
                        f"Configuration {k} is not a valid client "
                        f"configuration option. Valid client configuration options are "
                        f"{', '.join(k for k in valid_config_options)}."
                    )
                )
        self.__client_config = ClientConfig(**new_conf)

    def reset(self) -> None:
        """Reset LXD test environment provider to default configuration."""
        self.__configs = {
            InstanceConfig(name=name.replace("_", "-").lower(), source=source)
            for name, source in _DefaultSources.items()
        }
        self.__client_config = ClientConfig()
        super().reset()

    def add_instance_config(self, *new_config: InstanceConfig) -> None:
        """Add a new LXD instance configuration to the registry.

        Args:
            new_config (Dict[str, Any]):
                New configurations to add to the instance configuration registry.

        Raises:
            DuplicateLXDInstanceConfigError:
                Raised if two or more configs have the same name.
        """
        for config in new_config:
            if config.name not in [c.name for c in self.__configs]:
                self.__configs.add(config)
            else:
                raise DuplicateLXDInstanceConfigError(
                    f"Instance configuration with name {config.name} already exists."
                )

    def remove_instance_config(self, *name: str) -> None:
        """Remove an instance configuration by name.

        Do nothing if the instance configuration does not exist in the registry.

        Args:
            name (str): Names of the instance configurations to delete.
        """
        for config_name in name:
            for config in self.__configs:
                if config_name == config.name:
                    self.__configs.remove(config)

    def get_instance_config(self, name: str) -> InstanceConfig:
        """Return an LXD instance configuration.

        Args:
            name (str): Name of instance configuration to retrieve.

        Raises:
            LXDInstanceConfigNotFoundError:
                Raised if configuration does not exist in registry.

        Returns:
            (InstanceConfig): Retrieved LXD image configuration.
        """
        for config in self.__configs:
            if config.name == name:
                return copy.deepcopy(config)

        raise LXDInstanceConfigNotFoundError(f"Could not find instance {name}.")



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Hook run before test environment stops."""

from typing import List

from cleantest.meta import Injectable


class StopEnvHook:
    """Hook run before stopping test environment.

    Args:
        name (str): Unique name of hook.
        download (List[Injectable]): Artifacts to download from test environment.
    """

    def __init__(self, name: str = "default", download: List[Injectable] = []) -> None:
        self.name = name
        self.download = download



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Public metaclasses used inside cleantest."""

from .base_package import BasePackage, BasePackageError
from .injectable import Injectable
from .result import Result



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Hook run when test environment first starts."""

from typing import List

from cleantest.meta import Injectable


class StartEnvHook:
    """Hook run at the start of the test environment.

    Args:
        name (str): Unique name of hook.
        packages (List[Injectable]): Packages to inject into test environment.
        upload (List[Injectable]): Artifacts to upload into test environment.
    """

    def __init__(
        self,
        name: str = "default",
        packages: List[Injectable] = [],
        upload: List[Injectable] = [],
    ) -> None:
        self.name = name
        self.packages = packages
        self.upload = upload



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Detect operating system of test environment."""

import platform


class UnknownOSError(Exception):
    """Raised when an unknown operating system is detected."""


def detect_os_variant() -> str:
    """Detect the operating system of a test environment.

    Raises:
        UnknownOSError: Raised if an unknown operating system has been detected.

    Returns:
        (str): ID of the test environment's operating system.
    """
    try:
        dispatch = {
            "linux": _determine_linux,
            "darwin": lambda: "darwin",
            "windows": lambda: "windows",
            "java": lambda: "java",
        }
        return dispatch[platform.system().lower()]()
    except KeyError:
        raise UnknownOSError("Could not determine base operating system.")


def _determine_linux() -> str:
    """Determine Linux distribution by reading /etc/os-release.

    Returns:
        (str): ID of the Linux distribution read from /etc/os-release.
    """
    os_release_data = [line.strip() for line in open("/etc/os-release", "rt")]
    for line in os_release_data:
        if line.startswith("ID="):
            return line.split("=")[-1].lower()



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Common operations needed by classes to support snapd and snap."""

from shutil import which

from cleantest.meta.utils import detect_os_variant
from cleantest.utils import apt


class SnapdSupportError(Exception):
    """Base error for SnapdSupport mixin."""


class SnapdSupport:
    """Mixin for classes that need snapd support."""

    @staticmethod
    def _install_snapd() -> None:
        """Install snapd inside test environment.

        Raises:
            SnapdSupportError: Raised if snapd fails to install inside test environment.
            NotImplementedError: Raised if unsupported operating system is
                being used for a test environment.
        """
        os_variant = detect_os_variant()

        if which("snap") is None:
            if os_variant == "ubuntu":
                apt.install("snapd")
            else:
                raise NotImplementedError(
                    f"Support for {os_variant.capitalize()} not available yet."
                )



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Result class for cleantest methods that return data."""

from typing import Any


class Result:
    """Result containing exit code, stdout, and stderr.

    Args:
        exit_code (int): Captured exit code.
        stdout (Any): Captured data printed to standard output.
        stderr (Any): Captured data printed to standard error.
    """

    def __init__(self, exit_code: int, stdout: Any, stderr: Any):
        self.__exit_code = exit_code
        self.__stdout = stdout
        self.__stderr = stderr

    @property
    def exit_code(self) -> int:
        """Return captured exit_code."""
        return self.__exit_code

    @property
    def stdout(self) -> Any:
        """Return captured stdout."""
        return self.__stdout

    @property
    def stderr(self) -> Any:
        """Return captured stderr."""
        return self.__stderr

    def __repr__(self) -> str:
        """String representation of Result."""
        return (
            f"{self.__class__.__name__}(exit_code={self.__exit_code}, "
            f"stdout={self.__stdout}, stderr={self.__stderr})"
        )



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Manager for installing pip packages inside remote processes."""

import pathlib
import subprocess
import textwrap
from shutil import which
from typing import Dict, List, Union

from cleantest.meta import BasePackage, BasePackageError
from cleantest.meta.utils import detect_os_variant
from cleantest.utils import apt


class PipPackageError(BasePackageError):
    """Base error for Pip package handler."""


class Pip(BasePackage):
    """Represents `pip install`.

    Args:
        packages (List[str]): List of packages to install (Default: None).
        requirements (List[str]): List of paths to requirements.txt files (Default: None).
        constraints (List[str]): List of paths to constraints.txt files (Default: None).

    Raises:
        PipPackageError: Raised if lint rule fails at class creation.
    """

    def __init__(
        self,
        packages: Union[str, List[str]] = None,
        requirements: Union[str, List[str]] = None,
        constraints: Union[str, List[str]] = None,
    ) -> None:
        self.packages = [packages] if type(packages) == str else packages
        self.requirements = (
            [requirements] if type(requirements) == str else requirements
        )
        self._requirements_store = []
        self.constraints = [constraints] if type(constraints) == str else constraints
        self._constraints_store = []

        lint_rules = [
            lambda: True if packages is None and requirements is None else False,
            lambda: True if requirements is None and constraints is not None else False,
            lambda: True
            if type(requirements) == list
            and type(constraints) == list
            and len(requirements) != len(constraints)
            else False,
        ]
        for expr in lint_rules:
            if expr():
                raise PipPackageError(
                    "Lint rule failed. ",
                    f"Ensure passed arguments to {self.__class__.__name__} are correct.",
                )

    def _run(self) -> None:
        """Run Pip package handler."""
        self._setup()
        self._handle_pip_install()

    def _setup(self) -> None:
        """Set up and install dependencies needed by pip.

        Raises:
            PipPackageError: Raised if an error is encountered when installing pip.
            NotImplementedError: Raised if unsupported operating system is
                being used for a test environment.
        """
        os_variant = detect_os_variant()

        if which("pip") is None:
            if os_variant == "ubuntu":
                apt.install("python3-pip")
            else:
                raise NotImplementedError(
                    f"Support for {os_variant.capitalize()} not available yet."
                )

    def _handle_pip_install(self) -> None:
        """Install packages inside test environment using pip.

        Raises:
            PipPackageError: Raised if error is encountered when installing packages with pip.
        """
        if self.packages is not None:
            cmd = ["python3", "-m", "pip", "install", " ".join(self.packages)]
            try:
                subprocess.run(
                    cmd,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    check=True,
                )
            except subprocess.CalledProcessError:
                raise PipPackageError(
                    f"Failed to install packages {self.packages} "
                    f"using the following command {' '.join(cmd)}"
                )

        if self._constraints_store:
            for requirement, constraint in zip(
                self._requirements_store, self._constraints_store
            ):
                requirement_file = pathlib.Path.home().joinpath("requirements.txt")
                requirement_file.write_text(requirement)
                constraint_file = pathlib.Path.home().joinpath("constraints.txt")
                constraint_file.write_text(constraint)
                cmd = [
                    "python3",
                    "-m",
                    "pip",
                    "install",
                    "-r",
                    str(requirement_file),
                    "-c",
                    str(constraint_file),
                ]
                try:
                    subprocess.run(
                        cmd,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                        check=True,
                    )
                except subprocess.CalledProcessError:
                    raise PipPackageError(
                        (
                            f"Failed to install packages listed in requirements.txt file {requirement} "
                            f"with constraints.txt file {constraint} using the "
                            f"following command: {' '.join(cmd)}"
                        )
                    )
        else:
            for requirement in self._requirements_store:
                requirement_file = pathlib.Path(
                    pathlib.Path.home().joinpath("requirements.txt")
                )
                requirement_file.write_text(requirement)
                cmd = ["python3", "-m", "pip", "install", "-r", str(requirement_file)]
                try:
                    subprocess.run(
                        cmd,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                        check=True,
                    )
                except subprocess.CalledProcessError:
                    raise PipPackageError(
                        (
                            f"Failed to install packages listed in requirements.txt file {requirement} "
                            f"using the following command: {' '.join(cmd)}"
                        )
                    )

    def _dumps(self) -> Dict[str, str]:
        """Prepare Pip object for injection.

        Raises:
            FileNotFoundError: Raised if a requirements or constraints file is not found.

        Returns:
            (Dict[str, str]):
                checksum (str): Checksum to verify authenticity of serialized Pip object.
                data (str): Base64 encoded string containing serialized Pip object.
                injectable (str): Injectable to run inside remote environment.
        """
        if self.requirements is not None:
            for requirement in self.requirements:
                fin = pathlib.Path(requirement)
                if not fin.exists() or not fin.is_file():
                    raise FileNotFoundError(
                        f"Could not find requirements file {requirement}."
                    )
                self._requirements_store.append(fin.read_text())

        if self.constraints is not None:
            for constraint in self.constraints:
                fin = pathlib.Path(constraint)
                if not fin.exists() or not fin.is_file():
                    raise FileNotFoundError(
                        f"Could not find constraints file {constraint}."
                    )
                self._constraints_store.append(fin.read_text())

        return super()._dumps()

    def _injectable(self, data: Dict[str, str], **kwargs) -> str:
        """Generate injectable script that will be used to install packages with pip.

        Args:
            data (Dict[str, str]): Data that needs to be in injectable script.
                - checksum (str): SHA224 checksum to verify authenticity of Pip object.
                - data (str): Base64 encoded Pip object to inject.

        Returns:
            (str): Injectable script.
        """
        return textwrap.dedent(
            f"""
            #!/usr/bin/env python3

            from {self.__module__} import {self.__class__.__name__}

            holder = {self.__class__.__name__}._loads("{data['checksum']}", "{data['data']}")
            holder._run()
            """
        ).strip("\n")



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Direct the LXD test environment provider and instances."""

import csv
import json
import logging
import os
import pathlib
import shlex
import shutil
from collections import namedtuple
from concurrent.futures import ProcessPoolExecutor
from io import StringIO
from ipaddress import IPv4Address, IPv6Address
from typing import Dict, List, Optional, Tuple, Union

from pylxd import Client

from cleantest.control.lxd._lxd_configurer import LXDConfigurer
from cleantest.data import Dir, File
from cleantest.meta import Result
from cleantest.meta._cleantest_info import CleantestInfo
from cleantest.meta.utils import thread_count
from cleantest.utils import run

logger = logging.getLogger(__name__)

# Metaclass to encapsulate work information sent to _add threads.
_AddWorkOrder = namedtuple(
    "_AddWorkOrder",
    ["name", "image", "provision_script", "resources"],
    defaults=[None, None, None, None],
)

# Metaclass to encapsulate work information sent to _execute threads.
_ExecuteWorkOrder = namedtuple(
    "_ExecuteWorkOrder",
    ["target", "command"],
    defaults=[None, None],
)


class LXDArchonError(Exception):
    """Raise when LXDArchon encounters an error."""


class LXDArchon:
    """Direct the LXD test environment provider via the LXD API socket."""

    __instances = set()

    def __new__(cls) -> "LXDArchon":
        """Create new LXDArchon instance.

        Returns:
            (LXDArchon): New LXDArchon instance.
        """
        if not hasattr(cls, f"_{cls.__name__}__instance"):
            cls.__instance = super(LXDArchon, cls).__new__(cls)
        return cls.__instance

    @property
    def __client(self) -> Client:
        """Establish connection to LXD API socket.

        Returns:
            (Client): Connection to LXD API socket.
        """
        return Client(**self.config.client_config.dict())

    @property
    def config(self) -> LXDConfigurer:
        """Get LXDConfigurer instance containing configuration information.

        Returns:
            (LXDConfigurer): LXDConfigurer instance.
        """
        return LXDConfigurer()

    def exists(self, target: str) -> bool:
        """Check that specified test environment instance exists.

        Args:
            target (str): Instance to determine existence of.

        Returns:
            (bool): True if instance exists; False if otherwise.
        """
        return self.__client.instances.exists(target)

    def add(
        self,
        name: Union[str, List[str]],
        image: str,
        provision_script: Optional[Union[str, os.PathLike]] = None,
        resources: Optional[List[Union[File, Dir]]] = None,
    ) -> None:
        """Add a new test environment instance to the LXD cluster.

        Args:
            name (Union[str, List[str]]):
                Name of new test environment instance to add to the cluster.
            image (str):
                Image to use for new test environment instance.
            provision_script (Optional[Union[bytes, os.PathLike, str]]):
                Provisioning script to run after new instance starts.
            resources (Optional[List[Union[File, Dir]]]):
                Resources to upload into the test environment instance after
                new instance starts. You should only upload resources that are
                required for provisioning the test environment instance.
        """
        names = [name] if type(name) == str else name
        resources = resources if resources is not None else []
        with ProcessPoolExecutor(max_workers=thread_count()) as pool:
            for result in pool.map(
                self._add,
                (
                    _AddWorkOrder(name, image, provision_script, resources)
                    for name in names
                ),
            ):
                self.__instances.add(result)

    def _add(self, work_order: _AddWorkOrder) -> str:
        """Sub-function for setting up new test environment instance.

        Args:
            work_order (_AddWorkOrder):
                Information needed to add a new test environment instance.
        """
        _ = self.config.get_instance_config(work_order.image)
        _.name = work_order.name
        self.__client.instances.create(_.dict(), wait=True)
        instance = self.__client.instances.get(work_order.name)
        # TODO: Need to modify the start function so that it does not
        #   progress until a LXD VM has been assigned a public address.
        instance.start(wait=True)

        logger.info("Before injection")
        instance.execute(shlex.split("mkdir -p /root/.init/cleantest"))
        for name, data in CleantestInfo().dumps():
            instance.files.put(
                f"/root/.init/cleantest/install_{name}",
                data["injectable"],
            )
            instance.execute(
                shlex.split(f"python3 /root/.init/cleantest/install_{name}")
            )
        logger.info("After injection")
        for resource in work_order.resources:
            self.push(
                work_order.name, data_obj=resource, username="root", groupname="root"
            )

        if work_order.provision_script is not None:
            logger.info(f"Executing provision script {work_order.provision_script}")
            self.push(
                work_order.name,
                data_obj=[File(work_order.provision_script, "/root/.init/provision")],
            )
            instance.execute(shlex.split("python3 /root/.init/provision"))

        return work_order.name

    def execute(self, target: Union[str, List[str]], command: str) -> Dict[str, Result]:
        """Execute commands on test environment instances.

        Args:
            target (Union[str, List[str]]):
                Test environment instances to execute commands inside of.
            command (str):
                Command to execute.
        """
        _ = {}
        targets = [target] if type(target) == str else target
        [self.exists(target) for target in targets]
        with ProcessPoolExecutor(max_workers=thread_count()) as pool:
            for name, result in pool.map(
                self._execute,
                (_ExecuteWorkOrder(target, command) for target in targets),
            ):
                assert result.exit_code == 0
                _.update({name: result})

        return _

    def _execute(self, work_order: _ExecuteWorkOrder) -> Tuple[str, Result]:
        """Sub-function for executing commands on test environment instances.

        Args:
            work_order (_ExecuteWorkOrder):
                Information needed to execute command on a test environment instance.

        Returns:
            (Tuple[str, Result]): Name of target instance and result of execution.
        """
        instance = self.__client.instances.get(work_order.target)
        _ = instance.execute(shlex.split(work_order.command))
        return work_order.target, Result(_.exit_code, _.stdout, _.stderr)

    def pull(  # noqa C901
        self,
        target: str,
        src: Optional[Union[str, os.PathLike]] = None,
        dest: Optional[Union[str, os.PathLike]] = None,
        data_obj: Optional[List[Union[Dir, File]]] = None,
        uid: Optional[int] = None,
        username: Optional[str] = None,
        gid: Optional[int] = None,
        groupname: Optional[str] = None,
        mode: Optional[oct] = None,
        overwrite: bool = False,
    ) -> None:
        """Pull object from test environment instance.

        Args:
            target (str):
                Instance to pull object from.
            src (Optional[Union[str, os.PathLike]]):
                File path of object to pull from test environment instance.
                (Default: None).
            dest (Optional[Union[str, os.PathLike]]):
                Destination to pull object to. (Default: None).
            data_obj (Optional[List[Union[Dir, File]]]):
                Cleantest data object(s) to pull. (Default: None).
            uid (Optional[int]):
                uid to set as owner of object on host. (Default: None).
            username (Optional[str]):
                User to set as owner of object on host. (Default: None).
            gid (Optional[int]):
                gid to set as group of object on host. (Default: None).
            groupname (Optional[str]):
                Group name to set as group of object on host. (Default: None).
            mode (Optional[oct]):
                Mode to set on object on host. (Default: None).
            overwrite (bool):
                Overwrite object if it already exists on host. (Default: False).

        Raises:
            LXDArchonError: Raised if error is encountered when pulling object.
        """
        _objects = (
            [data_obj]
            if isinstance(data_obj, File) or isinstance(data_obj, Dir)
            else data_obj
            if data_obj is not None
            else []
        )

        if not self.exists(target):
            raise LXDArchonError(
                f"Instance {target} does not exist. Cannot pull object."
            )
        if src is None and dest is None and data_obj is None:
            raise LXDArchonError(f"Nothing to pull from {target}.")
        if (src is not None and dest is None) or (src is None and dest is not None):
            raise LXDArchonError(
                f"src or dest not specified. (src: {src}, dest: {dest}),"
            )
        if uid is not None and username is not None:
            raise LXDArchonError("Please specify only uid or username, not both.")
        if gid is not None and groupname is not None:
            raise LXDArchonError("Please specify only gid or groupname, not both.")

        instance = self.__client.instances.get(target)
        instance.execute(shlex.split("mkdir -p /root/.pull"))

        if src is not None and dest is not None:
            if instance.execute(shlex.split(f"test -f  {src}")).exit_code == 0:
                _ = File(src, dest, overwrite=overwrite)
            elif instance.execute(shlex.split(f"test -d {src}")).exit_code == 0:
                _ = Dir(src, dest, overwrite=overwrite)
            else:
                raise LXDArchonError(f"{src} is not a file or directory.")

            _objects.append(_)

        for _ in _objects:
            instance.files.put("/root/.pull/load", _._dumps(mode="pull")["injectable"])
            result = json.loads(
                instance.execute(shlex.split("python3 /root/.pull/load")).stdout
            )
            placeholder = _.__class__._loads(result["checksum"], result["data"])
            placeholder.dump()
            if (
                uid is not None
                or username is not None
                or gid is not None
                or groupname is not None
            ):
                if placeholder.dest.is_dir():
                    for d in placeholder.dest.iterdir():
                        shutil.chown(d, user=uid or username, group=gid or groupname)
                else:
                    shutil.chown(
                        placeholder.dest, user=uid or username, group=gid or groupname
                    )
            if mode is not None:
                if placeholder.dest.is_dir():
                    for d in placeholder.dest.iterdir():
                        d.chmod(mode)
                else:
                    placeholder.chmod(mode)

    def push(  # noqa C901
        self,
        target: str,
        src: Optional[Union[str, os.PathLike]] = None,
        dest: Optional[Union[str, os.PathLike]] = None,
        data_obj: Optional[List[Union[Dir, File]]] = None,
        uid: Optional[int] = None,
        username: Optional[str] = None,
        gid: Optional[int] = None,
        groupname: Optional[str] = None,
        mode: Optional[oct] = None,
        overwrite: bool = False,
    ) -> None:
        """Push object into test environment instance.

        Args:
            target (str):
                Instance to push object to.
            src (Optional[Union[str, os.PathLike]]):
                File path of object to push. (Default: None).
            dest (Optional[Union[str, os.PathLike]]):
                Destination to push object to. (Default: None).
            data_obj (Optional[List[Union[Dir, File]]]):
                Cleantest data object(s) to push. (Default: None).
            uid (Optional[int]):
                uid to set as owner of object inside test environment instance.
                (Default: None).
            username (Optional[str]):
                User to set as owner of object inside test environment instance.
                (Default: None).
            gid (Optional[int]):
                gid to set as group of object inside test environment instance.
                (Default: None).
            groupname (Optional[str]):
                Group name to set as group of object inside test environment instance.
                (Default: None).
            mode (Optional[oct]):
                Mode to set on object inside test environment instance.
                (Default: None).
            overwrite (bool):
                Overwrite object if it already exists in the test environment instance.
                (Default: False).

        Raises:
            LXDArchonError: Raised if error is encountered when pushing object.
        """
        _objects = (
            [data_obj]
            if isinstance(data_obj, File) or isinstance(data_obj, Dir)
            else data_obj
            if data_obj is not None
            else []
        )

        if not self.exists(target):
            raise LXDArchonError(
                f"Instance {target} does not exist. Cannot push object."
            )
        if src is None and dest is None and data_obj is None:
            raise LXDArchonError(f"Nothing to push to {target}.")
        if (src is not None and dest is None) or (src is None and dest is not None):
            raise LXDArchonError(
                f"src or dest not specified. (src: {src}, dest: {dest}),"
            )
        if uid is not None and username is not None:
            raise LXDArchonError("Please specify only uid or username, not both.")
        if gid is not None and groupname is not None:
            raise LXDArchonError("Please specify only gid or groupname, not both.")

        instance = self.__client.instances.get(target)
        instance.execute(shlex.split("mkdir -p /root/.push"))
        if src is not None and dest is not None:
            if pathlib.Path(src).is_file():
                _ = File(src, dest, overwrite=overwrite)
            elif pathlib.Path(src).is_dir():
                _ = Dir(src, dest, overwrite=overwrite)
            else:
                raise FileNotFoundError(f"Cannot locate src {src} on file system.")

            _objects.append(_)

        for _ in _objects:
            _.load()
            instance.files.put("/root/.push/dump", _._dumps(mode="push")["injectable"])
            instance.execute(shlex.split("python3 /root/.push/dump"))
            if (
                uid is not None
                or username is not None
                or gid is not None
                or groupname is not None
            ):
                instance.execute(
                    shlex.split(
                        f"chown -R {uid or username}:{gid or groupname} {str(_.dest)}"
                    )
                )
            if mode is not None:
                instance.execute(shlex.split(f"chmod -R {mode} {str(_.dest)}"))

    def destroy(self) -> None:
        """Destroy all test environment instances in LXD cluster."""
        logger.info(f"Destroying instances {self.__instances}")
        for i in self.__instances:
            instance = self.__client.instances.get(i)
            instance.stop(wait=True)
            instance.delete(wait=True)

    def get_public_address(
        self, target: str, ipv6: bool = False
    ) -> Optional[Union[IPv4Address, IPv6Address]]:
        """Get the public address of a test environment instance.

        Args:
            target (str): Name of test environment instance.
            ipv6 (bool): Get the public IPv6 address rather than IPv4 address.
                (Default: False).

        Returns:
            (Union[IPv4Address, IPv6Address]):
                Public address of test environment instance.
                None if the instance does not have a public address.
        """
        _ = csv.reader(StringIO(next(run("lxc list -c n46 -f csv")).stdout))
        for row in _:
            if row[0] == target and ipv6:
                return IPv6Address(row[2].split(" ")[0]) if row[2] != "" else None
            elif row[0] == target:
                return IPv4Address(row[1].split(" ")[0]) if row[1] != "" else None



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Classes for configuring the LXD test environment provider."""

from .lxd_config import ClientConfig, InstanceConfig, InstanceSource



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Metaclass for retrieving information about cleantest."""

import base64
import csv
import hashlib
import os
import pathlib
import tarfile
import tempfile
from concurrent.futures import ProcessPoolExecutor
from typing import Dict, Iterable, Tuple

import pkg_resources

from cleantest.meta.utils import thread_count


def _dependency_processor(dependency: pkg_resources.Distribution) -> Dict[str, bytes]:
    """Collect source code of cleantest dependency.

    Args:
        dependency (pkg_resources.Distribution): Dependency of cleantest.

    Returns:
        (Dict[str, bytes]): Name and base64 encoded source code of dependency.
    """
    os.chdir(dependency.location)
    with tempfile.NamedTemporaryFile() as fin:
        with tarfile.open(fin.name, "w:gz") as tar, pathlib.Path(
            f"{dependency.key.replace('-', '_')}-{dependency.version}.dist-info"
        ).joinpath("RECORD").open(mode="rt") as dist_info_fin:
            for row in csv.reader(dist_info_fin):
                tar.add(row[0])

        return {dependency.key: pathlib.Path(fin.name).read_bytes()}


class CleantestInfo:
    """Metaclass for getting information about the cleantest library."""

    def __new__(cls) -> "CleantestInfo":
        """Create new CleantestInfo object instance.

        Returns:
            (CleantestInfo): New object instance.
        """
        if not hasattr(cls, f"_{cls.__name__}__instance"):
            cls.__instance = super(CleantestInfo, cls).__new__(cls)
        return cls.__instance

    @property
    def __src(self) -> Dict[str, bytes]:
        """Retrieve the source code of cleantest.

        Returns:
            (Dict[str, bytes]): Name and base64 encoded source code of cleantest module.
        """
        _ = os.getcwd()
        os.chdir(pkg_resources.get_distribution("cleantest").location)
        with tempfile.NamedTemporaryFile() as fin:
            with tarfile.open(fin.name, "w:gz") as tar:
                tar.add("cleantest")
            os.chdir(_)
            return {"cleantest": pathlib.Path(fin.name).read_bytes()}

    @property
    def __dependencies(self) -> Iterable[Tuple[str, bytes]]:
        """Retrieve the source code of cleantest's dependencies.

        Yields:
            (Dict[str, bytes]): Name and source code of dependencies.
        """
        with ProcessPoolExecutor(max_workers=thread_count()) as pool:
            pool_results = pool.map(
                _dependency_processor,
                pkg_resources.working_set.resolve(
                    pkg_resources.working_set.by_key["cleantest"].requires()
                ),
            )
            for res in pool_results:
                for k, v in res.items():
                    yield k, v

    def __injectable(self, checksum: str, data: str) -> str:
        """Generate injectable script to install packages inside the test instance.

        Args:
            checksum (str): Checksum to verify base64 encoded object.
            data (str): Base64 encode tar archive containing source code for cleantest.

        Returns:
            (str): Injectable script.
        """
        with tempfile.TemporaryFile(mode="w+t") as fout:
            fout.writelines(
                [
                    "#!/usr/bin/env python3\n",
                    "import base64\n",
                    "import hashlib\n",
                    "import site\n",
                    "import tarfile\n",
                    "from io import BytesIO\n",
                    f"_ = base64.b64decode('{data}')\n",
                    f"if '{checksum}' != hashlib.sha224(_).hexdigest():\n"
                    "\traise Exception('Hashes do not match')\n",
                    "tar = tarfile.open(fileobj=BytesIO(_), mode='r:gz')\n",
                    "tar.extractall(site.getsitepackages()[0])\n",
                    "tar.close()\n",
                ]
            )
            fout.seek(0)
            return fout.read()

    def dumps(self) -> Iterable[Tuple[str, Dict[str, str]]]:
        """Prepare cleantest for injection into test environment instance.

        Yields:
            (Iterable[Tuple[str, Dict[str, str]]]):
                name (str): Name of library being injected.
                checksum (str): Checksum to verify authenticity of archive.
                data (str): Base64 encoded tarball containing source code.
                injectable (str): Injectable to run inside remote instance.
        """
        packages = self.__src
        packages.update(dict(self.__dependencies))
        for k, v in packages.items():
            checksum = hashlib.sha224(v).hexdigest()
            data = base64.b64encode(v).decode()
            yield k, {
                "checksum": checksum,
                "data": data,
                "injectable": self.__injectable(checksum, data),
            }



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Handler for LXD test environment provider and instances."""

import inspect
import json
import re
import tempfile
from concurrent.futures import ProcessPoolExecutor
from typing import Any, Callable, Iterable, List, Literal, Optional, Tuple

from pylxd import Client

from cleantest.meta import BasePackage, Injectable, Result
from cleantest.meta._base_harness import (
    BaseEntrypoint,
    BaseEntrypointError,
    BaseHarness,
)
from cleantest.meta._cleantest_info import CleantestInfo


class LXDEntrypointError(BaseEntrypointError):
    """Raise if error is encountered when starting test run with LXD."""


class InstanceMetadata:
    """Metaclass to track key information about LXD test environments.

    Args:
        name (str): Name of the test environment instance.
        image (Optional[str]):
            Name of image being used for the test environment instance
            (Default: None).
        exists (bool): Bool representing whether test environment exists.
            True - test environment instance exists.
            False - test environment instance does not exist (Default: False).
    """

    def __init__(
        self, name: str, image: Optional[str] = None, exists: bool = False
    ) -> None:
        self.name = name
        self.image = image
        self.exists = exists

    def __repr__(self) -> str:
        """String representation of InstanceMetadata."""
        return (
            f"{self.__class__.__name__}(name={self.name}, "
            f"image={self.image}, exists={self.exists})"
        )


class LXDHarness(BaseHarness):
    """Mixin for controlling the LXD hypervisor via its unix socket."""

    @property
    def _client(self) -> Client:
        """Get the connection to the LXD API socket.

        Returns:
            (Client): Connection to LXD API socket.
        """
        return Client(**self._lxd_config.client_config.dict())

    @property
    def _instance_metadata(self) -> List[InstanceMetadata]:
        """Create metaclasses to track key information about LXD test environments.

        Returns:
            (List[InstanceMetadata]): List of metaclasses.
        """
        return [
            InstanceMetadata(name=f"{self._name}-{i}", image=i) for i in self._image
        ]

    def _exists(self, instance: InstanceMetadata) -> InstanceMetadata:
        """Check whether an instance exists.

        Args:
            instance (InstanceMetadata): Instance to check the existence of.

        Returns:
            (InstanceMetadata): Update instance metadata.
        """
        if self._client.instances.exists(instance.name):
            instance.exists = True

        return instance

    def _init(self, instance: InstanceMetadata) -> None:
        """Initialize LXD test environment instance.

        Args:
            instance (InstanceMetadata): Instance to initialize.
        """
        if instance.exists is False:
            config = self._lxd_config.get_instance_config(instance.image)
            config.name = instance.name
            self._client.instances.create(config.dict(), wait=True)
            instance = self._client.instances.get(instance.name)
            instance.start(wait=True)
            meta = CleantestInfo()
            instance.execute(["mkdir", "-p", "/root/init/cleantest"])
            for name, data in meta.dumps():
                instance.files.put(
                    f"/root/init/cleantest/install_{name}",
                    data["injectable"],
                )
                instance.execute(["python3", f"/root/init/cleantest/install_{name}"])
        else:
            if self._client.instances.get(instance.name).status.lower() == "stopped":
                self._client.instances.get(instance.name).start(wait=True)

    def _execute(self, test: str, instance: InstanceMetadata) -> Any:
        """Execute a testlet inside an LXD test environment instance.

        Args:
            test (str): Testlet to execute inside test environment instance.
            instance (InstanceMetadata): Test environment instance to execute testlet inside.

        Returns:
            (Any): Result of the testlet.
        """
        instance = self._client.instances.get(instance.name)
        instance.files.put("/root/test", test)
        instance.execute(["chmod", "+x", "/root/test"])
        result = instance.execute(["/root/test"], environment=self._env.dump())
        return Result(
            exit_code=result.exit_code, stdout=result.stdout, stderr=result.stderr
        )

    def _teardown(self, instance: InstanceMetadata) -> None:
        """Teardown an LXD test environment instance after the testing has completed.

        Args:
            instance (InstanceMetadata): Test environment instance to teardown.
        """
        instance = self._client.instances.get(instance.name)
        instance.stop(wait=True)
        instance.delete(wait=True)

    def _handle_start_env_hooks(self, instance: InstanceMetadata) -> None:
        """Handle start env hooks.

        Args:
            instance (InstanceMetadata): Instance to run start env hooks in.
        """
        start_env_hooks = self._lxd_config.startenv_hooks
        while start_env_hooks:
            hook = start_env_hooks.pop()
            instance = self._client.instances.get(instance.name)
            if hook.packages is not None:
                for pkg in hook.packages:
                    self._handle_package_install(instance, pkg)
            if hook.upload is not None:
                for artifact in hook.upload:
                    self._handle_artifact_upload(instance, artifact)

    def _handle_stop_env_hooks(self, instance: InstanceMetadata) -> None:
        """Handle stop env hooks.

        Args:
            instance (InstanceMetadata): Instance to run stop env hooks in.
        """
        stop_env_hooks = self._lxd_config.stopenv_hooks
        while stop_env_hooks:
            hook = stop_env_hooks.pop()
            instance = self._client.instances.get(instance.name)
            if hook.download is not None:
                for artifact in hook.download:
                    self._handle_artifact_download(instance, artifact)

    def _handle_package_install(self, instance: Any, pkg: BasePackage) -> None:
        """Install a package inside an LXD test environment instance.

        Args:
            instance (Any): Instance to install package in.
            pkg (BasePackage): Package to install in instance.
        """
        dispatch = {"charmlib": lambda x: self._env.add(json.loads(x))}

        dump_data = pkg._dumps()
        instance.execute(["mkdir", "-p", "/root/init/pkg"])
        instance.files.put("/root/init/pkg/install", dump_data["injectable"])
        result = instance.execute(["python3", "/root/init/pkg/install"])

        if pkg.__class__.__name__.lower() in dispatch:
            dispatch[pkg.__class__.__name__.lower()](result.stdout)

    def _handle_artifact_upload(self, instance: Any, artifact: Injectable) -> None:
        """Upload an artifact to an LXD test environment instance.

        Args:
            instance (Any): Instance to upload artifact to.
            artifact (Injectable): Artifact to upload.
        """
        artifact.load()
        dump_data = artifact._dumps(mode="push")
        instance.execute(["mkdir", "-p", "/root/init/data"])
        instance.files.put("/root/init/data/dump", dump_data["injectable"])
        instance.execute(["python3", "/root/init/data/dump"])

    def _handle_artifact_download(self, instance: Any, artifact: Injectable) -> None:
        """Download an artifact from an LXD test environment instance.

        Args:
            instance (Any): Instance to download artifact from.
            artifact (Injectable): Artifact to download.
        """
        dump_data = artifact._dumps(mode="pull")
        instance.execute(["mkdir", "-p", "/root/post/data"])
        instance.files.put(
            "/root/post/data/load",
            dump_data["injectable"],
        )
        result = json.loads(
            instance.execute(["python3", "/root/post/data/load"]).stdout
        )
        with tempfile.NamedTemporaryFile():
            holder = artifact.__class__._loads(result["checksum"], result["data"])
            holder.dump()


class LXDProviderEntrypoint(BaseEntrypoint, LXDHarness):
    """Entrypoint to running testlets with LXD test environment provider.

    Args:
        strategy (Literal["serial", "serial_target", "parallel", "parallel_target"]):
        func (Callable):
        **kwargs (Any):
    """

    def __init__(
        self,
        strategy: Literal["serial", "serial_target", "parallel", "parallel_target"],
        func: Callable,
        **kwargs: Any,
    ) -> None:
        strategy_options = {
            "serial": self._serial_entrypoint,
            "serial_target": self._serial_target_entrypoint,
            "parallel": self._parallel_entrypoint,
            "parallel_target": self._parallel_target_entrypoint,
        }
        if strategy not in strategy_options.keys():
            raise LXDEntrypointError(
                (
                    f"{strategy} is not a valid strategy. "
                    f"Your options are {strategy_options}."
                )
            )
        setattr(self, "run", strategy_options[strategy])
        [setattr(self, k, v) for k, v in kwargs.items()]
        self._func = inspect.getsource(func)
        self._func_name = func.__name__

    def run(self) -> Iterable[Tuple[str, Result]]:
        """Method behavior is defined by passed strategy."""

    def _serial_entrypoint(self) -> Iterable[Tuple[str, Result]]:
        """Run testlets in serial. LXD instances will be created as needed.

        Yields:
            (Iterable[Tuple[str, Result]]):
                Aggregated results of testlet runs from each instance.
        """
        for instance in self._instance_metadata:
            yield self._run(instance)

    def _serial_target_entrypoint(self) -> Iterable[Tuple[str, Result]]:
        """Run testlets in serial inside pre-existing test environment instance.

        Yields:
            (Iterable[Tuple[str, Result]]):
                Aggregated results of testlet runs from each instance.
        """
        for instance in [
            self._exists(InstanceMetadata(name=name)) for name in self._target_instances
        ]:
            self._exists(instance)
            if not instance.exists:
                raise LXDEntrypointError(f"Instance {instance.name} does not exist.")
            yield self._run_target(instance)

    def _parallel_entrypoint(self) -> Iterable[Tuple[str, Result]]:
        """Run testlets in parallel inside pre-existing test environment instance.

        Yields:
            (Iterable[Tuple[str, Result]]):
                Aggregated results of testlet runs from each instance.
        """
        with ProcessPoolExecutor(max_workers=self._num_threads) as pool:
            results = pool.map(self._run, self._instance_metadata)
            for result in results:
                yield result

    def _parallel_target_entrypoint(self) -> Iterable[Tuple[str, Result]]:
        """Run testlets in parallel. LXD instances already exist.

        Yields:
            (Iterable[Tuple[str, Result]]):
                Aggregated results of testlet runs from each instance.
        """
        instance_metadata = [
            self._exists(InstanceMetadata(name=name)) for name in self._target_instances
        ]
        for instance in instance_metadata:
            if not instance.exists:
                raise LXDEntrypointError(f"Instance {instance.name} does not exist.")
        with ProcessPoolExecutor(max_workers=self._num_threads) as pool:
            results = pool.map(self._run_target, instance_metadata)
            for result in results:
                yield result

    def _run(self, instance: InstanceMetadata) -> Tuple[str, Result]:
        """Run testlet inside of test environment instance.

        Args:
            instance (InstanceMetadata):
                Test environment instance to operate on.

        Returns:
            (Tuple[str, Result]):
                Result of test run inside LXD test environment instance.
        """
        self._init(self._exists(instance))
        self._handle_start_env_hooks(instance)
        result = self._execute(
            self._make_testlet(
                self._func, self._func_name, [re.compile(r"^@lxd\(([^)]+)\)")]
            ),
            instance,
        )
        self._handle_stop_env_hooks(instance)
        if self._preserve is False:
            self._teardown(instance)

        return instance.name, result

    def _run_target(self, instance: InstanceMetadata) -> Tuple[str, Result]:
        """Run testlet inside pre-existing test environment instance.

        Args:
            instance (InstanceMetadata):
                Test environment instance to operate on.

        Returns:
            (Tuple[str, Result]):
                Result of test run inside of pre-existing
                LXD test environment instance.
        """
        self._handle_start_env_hooks(instance)
        result = self._execute(
            self._make_testlet(
                self._func, self._func_name, [re.compile(r"^@lxd\.target\(([^)]+)\)")]
            ),
            instance,
        )
        self._handle_stop_env_hooks(instance)
        return instance.name, result



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone, Canonical Ltd.
# See LICENSE file for licensing details.

from .charmlib import Charmlib
from .pip import Pip
from .snap import Alias, Confinement, Connection, Plug, Slot, Snap



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Utilities for testlets and cleantest."""

from . import apt
from . import snap
from . import systemd
from .run import run



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Dataclasses to assist with LXD provider and instance configuration."""

from typing import Dict, List, Optional, Tuple, Union

from cleantest.meta.mixins import DictLike, EnhancedEnum


class BadLXDConfigError(Exception):
    """Raised when the newly entered configuration fails the lint check."""


class ClientConfig(DictLike):
    """Define an LXD client connection.

    Args:
        endpoint (Optional[str]): Endpoint can be an HTTP endpoint or
            a path to a unix socket (Default: None).
        version (str): API version string to use with LXD
        cert (Optional[Tuple[str, str]]): A tuple of (cert, key) to use with
            the HTTP socket for client authentication (Default: "1.0").
        verify (bool): Either a boolean, in which case it controls
            whether we verify the server's TLS certificate, or a string, in
            which case it must be a path to a CA bundle to use.
            (Default: True).
        timeout (Optional[Union[float, Tuple[float, float]]]):
            How long to wait for the server to send data before giving up, as a float,
            or a (connect timeout, read timeout) tuple.
        project (Optional[str]): Name of the LXD project to interact with (Default: None).
    """

    def __init__(
        self,
        endpoint: Optional[str] = None,
        version: str = "1.0",
        cert: Optional[Tuple[str, str]] = None,
        verify: bool = True,
        timeout: Optional[Union[float, Tuple[float, float]]] = None,
        project: Optional[str] = None,
    ) -> None:
        self.endpoint = endpoint
        self.version = version
        self.cert = cert
        self.verify = verify
        self.timeout = timeout
        self.project = project

    def __repr__(self) -> str:
        """String representation of ClientConfig."""
        attrs = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{self.__class__.__name__}({attrs})"


class InstanceSource:
    """Define an LXD instance source to use for a test environment instance.

    Args:
        alias (str): Alias of source.
        mode (str): Mode for accessing source.
        protocol (str): Protocol to use when pulling source.
        server (str): Server to get source from.
        type (str): Type of source.
        allow_inconsistent (bool):
            Whether to ignore errors when copying. e.g. for volatile files. (Default: None).
        base_image (str):
            Base image fingerprint for faster migration. (Default: None).
        certificate (str):
            Certificate for remote images or migration. (Default: None).
        fingerprint (str): Fingerprint of image source. (Default: None).
        instance_only (bool):
            Whether the copy should skip the snapshots for copy. (Default: None).
        live (bool): Whether this is a live migration. (Default: None).
        operation (str):
            Remote operation URL for migration. (Default: None).
        project (str):
            Source project name for copy and local image. (Default: None).
        properties (Dict[str, str]): Image filters for image source. (Default: None).
        refresh (bool):
            Whether this is refreshing an existing instance for migration and copy.
            (Default: None).
        secret (str):
            Remote server secret for remote private images. (Default: None).
        secrets (Dict[str, str]):
            Map of migration websockets for migration. (Default: None).
        source (str):
            Existing instance name or snapshot for copy. (Default: None)
    """

    def __init__(
        self,
        alias: str,
        mode: str,
        protocol: str,
        server: str,
        type: str,
        allow_inconsistent: bool = None,
        base_image: str = None,
        certificate: str = None,
        fingerprint: str = None,
        instance_only: bool = None,
        live: bool = None,
        operation: str = None,
        project: str = None,
        properties: Dict[str, str] = None,
        refresh: bool = None,
        secret: str = None,
        secrets: Dict[str, str] = None,
        source: str = None,
    ) -> None:
        self.alias = alias
        self.type = type
        self.mode = mode
        self.server = server
        self.protocol = protocol
        if allow_inconsistent is not None:
            setattr(self, "allow_inconsistent", allow_inconsistent)
        if base_image is not None:
            setattr(self, "base_image", base_image)
        if certificate is not None:
            setattr(self, "certificate", certificate)
        if fingerprint is not None:
            setattr(self, "fingerprint", fingerprint)
        if instance_only is not None:
            setattr(self, "instance_only", instance_only)
        if live is not None:
            setattr(self, "live", live)
        if operation is not None:
            setattr(self, "operation", operation)
        if project is not None:
            setattr(self, "project", project)
        if properties is not None:
            setattr(self, "properties", properties)
        if refresh is not None:
            setattr(self, "refresh", refresh)
        if secret is not None:
            setattr(self, "secret", secret)
        if secrets is not None:
            setattr(self, "secrets", secrets)
        if source is not None:
            setattr(self, "source", source)

    def __repr__(self) -> str:
        """String representation of InstanceSource."""
        attrs = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{self.__class__.__name__}({attrs})"


class InstanceConfig(DictLike):
    """Define an LXD instance that can be brought up for test environments.

    Args:
        name (str): Name to use for container or virtual machine.
        source (InstanceSource): Source for LXD instance and where to get it from.
        architecture (str): Architecture name.
        config (Dict[str, str]): Instance configuration.
        description (str): Description of instance.
        devices (Dict[str, Dict[str, str]]): Devices to use with instance.
        ephemeral (bool):
            Whether the instance is ephemeral; deleted on shutdown. (Default: False).
        instance_type (str):
            Cloud instance type (AWS, GCP, Azure, ...) to emulate with limits. (Default: None).
        profiles (List[str]):
            List of profiles apply to the instance. (Default: None).
        restore (str):
            If set, instance will be restored to the provided snapshot name. (Default: None).
        stateful (bool):
            Whether the instance currently has saved state on the host. (Default: False).
        type (str):
            Type of instance. i.e. "container" or "vm". (Default: "container").
    """

    def __init__(
        self,
        name: str,
        source: InstanceSource,
        architecture: str = None,
        config: Dict[str, str] = None,
        description: str = None,
        devices: Dict[str, Dict[str, str]] = None,
        ephemeral: bool = None,
        instance_type: str = None,
        profiles: List[str] = None,
        restore: str = None,
        stateful: bool = None,
        type: str = "container",
    ) -> None:
        self.name = name
        self.source = source
        self.type = type
        if architecture is not None:
            setattr(self, "architecture", architecture)
        if config is not None:
            setattr(self, "config", config)
        if description is not None:
            setattr(self, "description", description)
        if devices is not None:
            setattr(self, "devices", devices)
        if ephemeral is not None:
            setattr(self, "ephemeral", ephemeral)
        if instance_type is not None:
            setattr(self, "instance_type", instance_type)
        if profiles is not None:
            setattr(self, "profiles", profiles)
        if restore is not None:
            setattr(self, "restore", restore)
        if stateful is not None:
            setattr(self, "stateful", stateful)

        self.__lint()

    def __lint(self) -> None:
        """Lint a new LXD instance configuration to ensure that it is valid.

        Raises:
            BadLXDConfigError:
                Raised if the passed LXD instance configuration is invalid.
        """
        checks = ["name", "server", "alias", "protocol", "type", "mode"]
        for i in checks:
            if i not in self.keys(all_keys=True):
                raise BadLXDConfigError(
                    (
                        f"Bad instance configuration: {self.dict()}. "
                        "Please ensure instance configuration has the "
                        f"following values set: {', '.join(checks)}."
                    )
                )

    def __repr__(self) -> str:
        """String representation of InstanceConfig."""
        attrs = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{self.__class__.__name__}({attrs})"


class _DefaultSources(EnhancedEnum):
    """Default sources for LXD test environment provider."""

    ALMALINUX_8_AMD64 = InstanceSource(
        alias="almalinux/8/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    ALMALINUX_8_ARM64 = InstanceSource(
        alias="almalinux/8/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    ALMALINUX_9_AMD64 = InstanceSource(
        alias="almalinux/9/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    ALMALINUX_9_ARM64 = InstanceSource(
        alias="almalinux/9/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    ARCHLINUX_AMD64 = InstanceSource(
        alias="archlinux/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    ARCHLINUX_ARM64 = InstanceSource(
        alias="archlinux/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    CENTOS_8_STREAM_AMD64 = InstanceSource(
        alias="centos/8-Stream/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    CENTOS_8_STREAM_ARM64 = InstanceSource(
        alias="centos/8-Stream/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    CENTOS_9_STREAM_AMD64 = InstanceSource(
        alias="centos/9-Stream/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    CENTOS_9_STREAM_ARM64 = InstanceSource(
        alias="centos/9-Stream/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    DEBIAN_10_AMD64 = InstanceSource(
        alias="debian/10/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    DEBIAN_10_ARM64 = InstanceSource(
        alias="debian/10/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    DEBIAN_11_AMD64 = InstanceSource(
        alias="debian/11/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    DEBIAN_11_ARM64 = InstanceSource(
        alias="debian/11/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    DEBIAN_12_AMD64 = InstanceSource(
        alias="debian/12/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    DEBIAN_12_ARM64 = InstanceSource(
        alias="debian/12/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    FEDORA_35_AMD64 = InstanceSource(
        alias="fedora/35/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    FEDORA_35_ARM64 = InstanceSource(
        alias="fedora/35/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    FEDORA_36_AMD64 = InstanceSource(
        alias="fedora/36/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    FEDORA_36_ARM64 = InstanceSource(
        alias="fedora/36/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    FEDORA_37_AMD64 = InstanceSource(
        alias="fedora/37/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    FEDORA_37_ARM64 = InstanceSource(
        alias="fedora/37/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    ROCKYLINUX_8_AMD64 = InstanceSource(
        alias="rockylinux/8/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    ROCKYLINUX_8_ARM64 = InstanceSource(
        alias="rockylinux/8/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    ROCKYLINUX_9_AMD64 = InstanceSource(
        alias="rockylinux/9/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    ROCKYLINUX_9_ARM64 = InstanceSource(
        alias="rockylinux/9/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    UBUNTU_JAMMY_AMD64 = InstanceSource(
        alias="ubuntu/jammy/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    UBUNTU_JAMMY_ARM64 = InstanceSource(
        alias="ubuntu/jammy/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    UBUNTU_FOCAL_AMD64 = InstanceSource(
        alias="ubuntu/focal/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    UBUNTU_FOCAL_ARM64 = InstanceSource(
        alias="ubuntu/focal/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    UBUNTU_BIONIC_AMD64 = InstanceSource(
        alias="ubuntu/18.04/amd64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )

    UBUNTU_BIONIC_ARM64 = InstanceSource(
        alias="ubuntu/18.04/arm64",
        mode="pull",
        protocol="simplestreams",
        server="https://images.linuxcontainers.org",
        type="image",
    )



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Common operations for Singletons that need to be resettable."""

from abc import ABC, abstractmethod


class Resettable(ABC):
    """Abstract mixin for Singletons that need to implement reset behavior."""

    @abstractmethod
    def reset(self) -> None:
        """Reset the current Singleton to its default state."""



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Handler for installing and managing snap packages inside a remote instance."""

import http.client
import json
import os
import re
import socket
import subprocess
import sys
import urllib.error
import urllib.parse
import urllib.request
from collections.abc import Mapping
from datetime import datetime, timedelta, timezone
from enum import Enum
from subprocess import CalledProcessError, CompletedProcess
from typing import Any, Callable, Dict, Iterable, List, Union

# Filter for extracting 7-bit C1 ANSI sequences from a string.
ansi_filter = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


class SnapHandlerError(Exception):
    """Raised when snap handler encounters any errors."""


class SnapAPIError(Exception):
    """Raised when an HTTP API error occurs talking to the Snapd server."""

    def __init__(self, body: Dict, code: int, status: str, message: str):
        """This shouldn't be instantiated directly."""
        super().__init__(message)  # Makes str(e) return message
        self.body = body
        self.code = code
        self.status = status
        self._message = message

    def __repr__(self):
        """String representation of the SnapAPIError class."""
        return f"APIError({self.body}, {self.code}, {self.status}, {self._message})"


class SnapState(Enum):
    """The state of a snap on the system or in the cache."""

    PRESENT = "present"
    ABSENT = "absent"
    LATEST = "latest"
    AVAILABLE = "available"


def _cache_init(func: Callable):
    def inner(*args, **kwargs):
        if _Cache.cache is None:
            _Cache.cache = SnapCache()
        return func(*args, **kwargs)

    return inner


class MetaCache(type):
    """MetaCache class used for initialising the snap cache."""

    @property
    def cache(cls) -> "SnapCache":
        """Property for returning the snap cache."""
        return cls._cache

    @cache.setter
    def cache(cls, cache: "SnapCache") -> None:
        """Setter for the snap cache."""
        cls._cache = cache

    def __getitem__(cls, name) -> "Snap":
        """Snap cache getter."""
        return cls._cache[name]


class _Cache(object, metaclass=MetaCache):
    _cache = None


class SnapService:
    """Data wrapper for snap services."""

    def __init__(
        self,
        daemon: str = None,
        daemon_scope: str = None,
        enabled: bool = False,
        active: bool = False,
        activators: List[str] = [],
        **kwargs,
    ):
        self.daemon = daemon
        self.daemon_scope = kwargs.get("daemon-scope", None) or daemon_scope
        self.enabled = enabled
        self.active = active
        self.activators = activators


class Snap:
    """Represents a snap package and its properties.

    `Snap` exposes the following properties about a snap:
      - name: the name of the snap
      - state: a `SnapState` representation of its install status
      - channel: "stable", "candidate", "beta", and "edge" are common
      - revision: a string representing the snap's revision
      - confinement: "classic" or "strict"
    """

    def __init__(
        self,
        name,
        state: SnapState,
        channel: str,
        revision: str,
        confinement: str,
        apps: List[Dict[str, str]] = None,
        cohort: str = "",
    ) -> None:
        self._name = name
        self._state = state
        self._channel = channel
        self._revision = revision
        self._confinement = confinement
        self._cohort = cohort
        self._apps = apps or []
        self._snap_client = SnapClient()

    def __eq__(self, other) -> bool:
        """Equality for comparison."""
        return isinstance(other, self.__class__) and (
            self._name,
            self._revision,
        ) == (other._name, other._revision)

    def __hash__(self):
        """A basic hash so this class can be used in Mappings and dicts."""
        return hash((self._name, self._revision))

    def __repr__(self):
        """A representation of the snap."""
        return f"<{self.__module__}.{self.__class__.__name__}: {self.__dict__}>"

    def __str__(self):
        """A human-readable representation of the snap."""
        return f"<{self.__class__.__name__}: {self._name}-{self._revision}.{self._channel} -- {str(self._state)}>"

    def _snap(self, command: str, optargs: Iterable[str] = None) -> str:
        """Perform a snap operation.

        Args:
            command (str): The snap command to execute
            optargs (Iterable[str]): A list of additional arguments to pass,
                commonly confinement or channel

        Raises:
          SnapHandlerError: Raised there is a problem encountered when executing snap command.

        Returns:
            (str): Captured output of operation performed on snap.
        """
        optargs = optargs or []
        _cmd = ["snap", command, self._name, *optargs]
        try:
            return subprocess.check_output(_cmd, universal_newlines=True)
        except CalledProcessError as e:
            raise SnapHandlerError(
                f"Snap: {self._name}; command {_cmd} failed with output = {e.output}"
            )

    def _snap_daemons(
        self,
        command: List[str],
        services: List[str] = None,
    ) -> CompletedProcess:
        """Execute command on a list of snap services.

        Args:
            command (List[str]): Command to execute on snap services.
            services (List[str]): Services to operate on (Default: None).

        Returns:
            (CompletedProcess): Captured output of operation on snap services.
        """
        if services:
            # an attempt to keep the command constrained to the snap instance's services
            services = [f"{self._name}.{service}" for service in services]
        else:
            services = [self._name]

        _cmd = ["snap", *command, *services]

        try:
            return subprocess.run(
                _cmd, universal_newlines=True, check=True, capture_output=True
            )
        except CalledProcessError as e:
            raise SnapHandlerError(
                f"Could not {_cmd} for snap [{self._name}]: {e.stderr}"
            )

    def get(self, key) -> str:
        """Gets a snap configuration value.

        Args:
            key (str): The key to retrieve.

        Returns:
            (str): Captured output of get operation.
        """
        return self._snap("get", [key]).strip()

    def set(self, config: Dict) -> str:
        """Sets a snap configuration value.

        Args:
           config (Dict): A dictionary containing keys and values specifying the config to set.

        Returns:
            (str): Captured output of set operation.
        """
        args = [f'{key}="{val}"' for key, val in config.items()]

        return self._snap("set", [*args])

    def unset(self, key: str) -> str:
        """Unsets a snap configuration value.

        Args:
            key (str): The key to unset.

        Returns:
            (str): Captured output of unset operation.
        """
        return self._snap("unset", [key])

    def start(self, services: List[str] = None, enable: bool = False) -> None:
        """Starts a snap's services.

        Args:
            services (List[str]): List of individual snap services to start,
                otherwise all (Default: None).
            enable (bool): Flag to enable snap services on start (Default: False).
        """
        args = ["start", "--enable"] if enable else ["start"]
        self._snap_daemons(args, services)

    def stop(self, services: List[str] = None, disable: bool = False) -> None:
        """Stops a snap's services.

        Args:
            services (List[str]): List of individual snap services to stop,
                otherwise all (Default: None).
            disable (bool): Flag to disable snap services on stop (Default: False).
        """
        args = ["stop", "--disable"] if disable else ["stop"]
        self._snap_daemons(args, services)

    def logs(self, services: List[str] = None, num_lines: int = 10) -> str:
        """Shows a snap services' logs.

        Args:
            services (List[str]): List of individual snap services to show logs from,
                otherwise all (Default: None).
            num_lines (int): Integer number of log lines to return (Default: 10).

        Returns:
            (str): Captured output of logs operation.
        """
        args = ["logs", f"-n={num_lines}"] if num_lines else ["logs"]
        return self._snap_daemons(args, services).stdout

    def restart(self, services: List[str] = None, reload: bool = False) -> None:
        """Restarts a snap's services.

        Args:
            services (list): List of individual snap services to restart,
                otherwise all (Default: None).
            reload (bool): Flag to use the service reload command, if available (Default: False).
        """
        args = ["restart", "--reload"] if reload else ["restart"]
        self._snap_daemons(args, services)

    def _install(self, channel: str = "", cohort: str = "") -> None:
        """Add a snap to the system.

        Args:
            channel (str): The channel to install from (Default: "").
            cohort (str): The key of a cohort that this snap belongs to (Default: "").
        """
        cohort = cohort or self._cohort

        args = []
        if self.confinement == "classic":
            args.append("--classic")
        if channel:
            args.append(f'--channel="{channel}"')
        if cohort:
            args.append(f'--cohort="{cohort}"')

        self._snap("install", args)

    def _refresh(
        self,
        channel: str = "",
        cohort: str = "",
        leave_cohort: bool = False,
    ) -> None:
        """Refresh a snap.

        Args:
            channel (str): The channel to install from (Default: "").
            cohort (str): Specify a cohort (Default: "").
            leave_cohort (bool): Leave the current cohort (Default: False).
        """
        channel = f'--channel="{channel}"' if channel else ""
        args = [channel]

        if not cohort:
            cohort = self._cohort

        if leave_cohort:
            self._cohort = ""
            args.append("--leave-cohort")
        elif cohort:
            args.append(f'--cohort="{cohort}"')

        self._snap("refresh", args)

    def _remove(self) -> str:
        """Removes a snap from the system.

        Returns:
            (str): Captured output of remove operation.
        """
        return self._snap("remove")

    @property
    def name(self) -> str:
        """Get the name of the snap.

        Returns:
            (str): Name of the snap.
        """
        return self._name

    def ensure(
        self,
        state: SnapState,
        classic: bool = False,
        channel: str = "",
        cohort: str = "",
    ):
        """Ensures that the snap is in a given state.

        Args:
            state (SnapState): A `SnapState` to reconcile to.
            classic (bool): A boolean indicating whether classic confinement should
                be used (Default: False).
            channel (str): The channel to install from (Default: "").
            cohort (str): Specify the key of a snap cohort (Default: "").

        Raises:
            SnapHandlerError: Raised if an error is encountered while
                ensuring that the snap is in a given state.
        """
        self._confinement = (
            "classic" if classic or self._confinement == "classic" else ""
        )

        if state not in (SnapState.PRESENT, SnapState.LATEST):
            # We are attempting to remove this snap.
            if self._state in (SnapState.PRESENT, SnapState.LATEST):
                # The snap is installed, so we run _remove.
                self._remove()
            else:
                # The snap is not installed -- no need to do anything.
                pass
        else:
            # We are installing or refreshing a snap.
            if self._state not in (SnapState.PRESENT, SnapState.LATEST):
                # The snap is not installed, so we install it.
                self._install(channel, cohort)
            else:
                # The snap is installed, but we are changing it (e.g., switching channels).
                self._refresh(channel, cohort)

        self._update_snap_apps()
        self._state = state

    def _update_snap_apps(self) -> None:
        """Updates a snap's apps after snap changes state."""
        try:
            self._apps = self._snap_client.get_installed_snap_apps(self._name)
        except SnapAPIError:
            self._apps = []

    @property
    def present(self) -> bool:
        """Get whether a snap is present.

        Returns:
            (bool): True - snap is present. False - snap is not present.
        """
        return self._state in (SnapState.PRESENT, SnapState.LATEST)

    @property
    def latest(self) -> bool:
        """Get whether the snap is the most recent version.

        Returns:
            (bool): True - snap is the most recent version.
                False - snap is not the most recent version.
        """
        return self._state is SnapState.LATEST

    @property
    def state(self) -> SnapState:
        """Get the current snap state.

        Returns:
            (SnapState): Current snap state.
        """
        return self._state

    @state.setter
    def state(self, state: SnapState) -> None:
        """Sets the snap state to a given value.

        Args:
          state (SnapState): A `SnapState` to reconcile the snap to.

        Raises:
          SnapHandlerError: Raised if an error is encountered while setting
            the state of the snap.
        """
        if self._state is not state:
            self.ensure(state)
        self._state = state

    @property
    def revision(self) -> str:
        """Get the revision of the snap.

        Returns:
            (str): Revision of the snap.
        """
        return self._revision

    @property
    def channel(self) -> str:
        """Get the channel of the snap.

        Returns:
            (str): Channel of the snap.
        """
        return self._channel

    @property
    def confinement(self) -> str:
        """Get the confinement of the snap.

        Returns:
            (str): Confinement of the snap.
        """
        return self._confinement

    @property
    def apps(self) -> List:
        """Get the installed apps of the snap.

        Returns:
            (List): Installed apps of the snap, if any.
        """
        self._update_snap_apps()
        return self._apps

    @property
    def services(self) -> Dict:
        """Get the installed services of the snap.

        Returns:
            (Dict): Installed services of the snap, if any.
        """
        self._update_snap_apps()
        services = {}
        for app in self._apps:
            if "daemon" in app:
                services[app["name"]] = SnapService(**app).__dict__

        return services


class _UnixSocketConnection(http.client.HTTPConnection):
    """Implementation of HTTPConnection that connects to a named Unix socket."""

    def __init__(
        self, host: str, timeout: float = None, socket_path: Any = None
    ) -> None:
        if timeout is None:
            super().__init__(host)
        else:
            super().__init__(host, timeout=timeout)
        self.socket_path = socket_path

    def connect(self) -> None:
        """Override connect to use Unix socket (instead of TCP socket)."""
        if not hasattr(socket, "AF_UNIX"):
            raise NotImplementedError(f"Unix sockets not supported on {sys.platform}")
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(self.socket_path)
        if self.timeout is not None:
            self.sock.settimeout(self.timeout)


class _UnixSocketHandler(urllib.request.AbstractHTTPHandler):
    """Implementation of HTTPHandler that uses a named Unix socket."""

    def __init__(self, socket_path: str):
        super().__init__()
        self.socket_path = socket_path

    def http_open(self, req) -> http.client.HTTPResponse:
        """Override http_open to use a Unix socket connection (instead of TCP)."""
        return self.do_open(_UnixSocketConnection, req, socket_path=self.socket_path)


class SnapClient:
    """Snapd API client to talk to HTTP over UNIX sockets.

    In order to avoid shelling out and/or involving sudo in calling the snapd API,
    use a wrapper based on the Pebble Client, trimmed down to only the utility methods
    needed for talking to snapd.
    """

    def __init__(
        self,
        socket_path: str = "/run/snapd.socket",
        opener: urllib.request.OpenerDirector = None,
        base_url: str = "http://localhost/v2/",
        timeout: float = 5.0,
    ):
        """Initialize a client instance.

        Args:
            socket_path (str): A path to the socket on the filesystem
                (Default: "/run/snap/snapd.socket").
            opener (urllib.request.OpenerDirector): Specifies an opener for unix socket.
                If set to None, a default is used.
            base_url (str): Base url for making requests to the snap client. Defaults to
                (Default: "http://localhost/v2/")
            timeout (float): Timeout in seconds to use when making requests to the API.
                (Default: 5.0).
        """
        if opener is None:
            opener = self._get_default_opener(socket_path)
        self.opener = opener
        self.base_url = base_url
        self.timeout = timeout

    @classmethod
    def _get_default_opener(cls, socket_path: str) -> urllib.request.OpenerDirector:
        """Build the default opener to use for requests (HTTP over Unix socket)."""
        opener = urllib.request.OpenerDirector()
        opener.add_handler(_UnixSocketHandler(socket_path))
        opener.add_handler(urllib.request.HTTPDefaultErrorHandler())
        opener.add_handler(urllib.request.HTTPRedirectHandler())
        opener.add_handler(urllib.request.HTTPErrorProcessor())
        return opener

    def _request(
        self,
        method: str,
        path: str,
        query: Dict = None,
        body: Dict = None,
    ) -> Union[dict, List]:
        """Make a JSON request to the Snapd server with the given HTTP method and path.

        If query dict is provided, it is encoded and appended as a query string
        to the URL. If body dict is provided, it is serialized as JSON and used
        as the HTTP body (with Content-Type: "application/json"). The resulting
        body is decoded from JSON.
        """
        headers = {"Accept": "application/json"}
        data = None
        if body is not None:
            data = json.dumps(body).encode("utf-8")
            headers["Content-Type"] = "application/json"

        response = self._request_raw(method, path, query, headers, data)
        return json.loads(response.read().decode())["result"]

    def _request_raw(
        self,
        method: str,
        path: str,
        query: Dict = None,
        headers: Dict = None,
        data: bytes = None,
    ) -> http.client.HTTPResponse:
        """Make a request to the Snapd server; return the raw HTTPResponse object."""
        url = self.base_url + path
        if query:
            url = url + "?" + urllib.parse.urlencode(query)

        if headers is None:
            headers = {}
        request = urllib.request.Request(url, method=method, data=data, headers=headers)

        try:
            response = self.opener.open(request, timeout=self.timeout)
        except urllib.error.HTTPError as e:
            code = e.code
            status = e.reason
            message = ""
            try:
                body = json.loads(e.read().decode())["result"]
            except (IOError, ValueError, KeyError) as e2:
                # Will only happen on read error or if Pebble sends invalid JSON.
                body = {}
                message = f"{type(e2).__name__} - {e2}"
            raise SnapAPIError(body, code, status, message)
        except urllib.error.URLError as e:
            raise SnapAPIError({}, 500, "Not found", e.reason)
        return response

    def get_installed_snaps(self) -> Dict:
        """Get information about currently installed snaps."""
        return self._request("GET", "snaps")

    def get_snap_information(self, name: str) -> Dict:
        """Query the snap server for information about single snap."""
        return self._request("GET", "find", {"name": name})[0]

    def get_installed_snap_apps(self, name: str) -> List:
        """Query the snap server for apps belonging to a named, currently installed snap."""
        return self._request("GET", "apps", {"names": name, "select": "service"})


class SnapCache(Mapping):
    """An abstraction to represent installed/available packages.

    When instantiated, `SnapCache` iterates through the list of installed
    snaps using the `snapd` HTTP API, and a list of available snaps by reading
    the filesystem to populate the cache. Information about available snaps is lazily-loaded
    from the `snapd` API when requested.
    """

    def __init__(self):
        if not self.snapd_installed:
            raise SnapHandlerError(
                "snapd is not installed or not in /usr/bin"
            ) from None
        self._snap_client = SnapClient()
        self._snap_map = {}
        if self.snapd_installed:
            self._load_available_snaps()
            self._load_installed_snaps()

    def __contains__(self, key: str) -> bool:
        """Magic method to ease checking if a given snap is in the cache."""
        return key in self._snap_map

    def __len__(self) -> int:
        """Returns number of items in the snap cache."""
        return len(self._snap_map)

    def __iter__(self) -> Iterable["Snap"]:
        """Magic method to provide an iterator for the snap cache."""
        return iter(self._snap_map.values())

    def __getitem__(self, snap_name: str) -> Snap:
        """Return either the installed version or latest version for a given snap."""
        snap = self._snap_map.get(snap_name, None)
        if snap is None:
            # The snapd cache file may not have existed when _snap_map was
            # populated.  This is normal.
            try:
                self._snap_map[snap_name] = self._load_info(snap_name)
            except SnapAPIError:
                raise SnapHandlerError(f"Snap '{snap_name}' not found!")

        return self._snap_map[snap_name]

    @property
    def snapd_installed(self) -> bool:
        """Check whether snapd has been installed on the system."""
        return os.path.isfile("/usr/bin/snap")

    def _load_available_snaps(self) -> None:
        """Load the list of available snaps from disk.

        Leave them empty and lazily load later if asked for.
        """
        if not os.path.isfile("/var/cache/snapd/names"):
            # The snap catalog may not be populated yet; this is normal.
            # snapd updates the cache infrequently and the cache file may not
            # currently exist.
            return

        with open("/var/cache/snapd/names", "r") as f:
            for line in f:
                if line.strip():
                    self._snap_map[line.strip()] = None

    def _load_installed_snaps(self) -> None:
        """Load the installed snaps into the dict."""
        installed = self._snap_client.get_installed_snaps()

        for i in installed:
            snap = Snap(
                name=i["name"],
                state=SnapState.LATEST,
                channel=i["channel"],
                revision=i["revision"],
                confinement=i["confinement"],
                apps=i.get("apps", None),
            )
            self._snap_map[snap.name] = snap

    def _load_info(self, name: str) -> Snap:
        """Load info for snaps which are not installed if requested.

        Args:
            name (str): A string representing the name of the snap.
        """
        info = self._snap_client.get_snap_information(name)

        return Snap(
            name=info["name"],
            state=SnapState.AVAILABLE,
            channel=info["channel"],
            revision=info["revision"],
            confinement=info["confinement"],
            apps=None,
        )


@_cache_init
def install(
    snap_names: Union[str, List[str]],
    state: Union[str, SnapState] = SnapState.LATEST,
    channel: str = "latest",
    classic: bool = False,
    cohort: str = "",
) -> Union[Snap, List[Snap]]:
    """Install a snap or snaps in a remote process.

    Args:
        snap_names (Union[str, List[str]]): The name or names of the snaps to install.
        state (SnapState): A string or `SnapState` representation of the desired state, one of
            [`PRESENT` or `LATEST`] (Default: SnapState.LATEST).
        channel (str): A channel as a string (Default: "latest").
        classic (bool): A boolean specifying whether it should be added with classic
            confinement (Default: False).
        cohort (str): The key of a cohort that this snap belongs to (Default: "").

    Raises:
        SnapHandlerError: Raised if some snaps failed to install or were not found.

    Returns:
        (Union[Snap, List[Snap]]): Result of operation.
    """
    snap_names = [snap_names] if type(snap_names) is str else snap_names
    if not snap_names:
        raise TypeError("Expected at least one snap to add, received zero!")

    if type(state) is str:
        state = SnapState(state)

    return _wrap_snap_operations(snap_names, state, channel, classic, cohort)


@_cache_init
def remove(snap_names: Union[str, List[str]]) -> Union[Snap, List[Snap]]:
    """Removes a snap from the system.

    Args:
        snap_names (Union[str, List[str]]): The name or names of the snaps to remove.

    Raises:
        SnapHandlerError: Raised if some snaps failed to be removed.

    Returns:
        (Union[Snap, List[Snap]]): Result of operation.
    """
    snap_names = [snap_names] if type(snap_names) is str else snap_names
    if not snap_names:
        raise TypeError("Expected at least one snap to add, received zero!")

    return _wrap_snap_operations(snap_names, SnapState.ABSENT, "", False)


@_cache_init
def ensure(
    snap_names: Union[str, List[str]],
    state: str,
    channel: str = "latest",
    classic: bool = False,
    cohort: str = "",
) -> Union[Snap, List[Snap]]:
    """Ensures a snap is in a given state to the system.

    Args:
        snap_names (Union[str, List[str]]): The name(s) of the snaps to operate on.
        state (str): A string representation of the desired state, from `SnapState`.
        channel (str): A channel as a string (Default: "latest").
        classic (bool): A boolean specifying whether it should be added with classic.
            confinement (Default: False).
        cohort (str): The key of a cohort that this snap belongs to (Default: "").

    Raises:
        SnapHandlerError: Raised if the snap is not in the cache.
    """
    if state in ("present", "latest"):
        return install(snap_names, SnapState(state), channel, classic, cohort)
    else:
        return remove(snap_names)


def _wrap_snap_operations(
    snap_names: List[str],
    state: SnapState,
    channel: str,
    classic: bool,
    cohort: str = "",
) -> Union[Snap, List[Snap]]:
    """Wrap common operations for bare snap commands.

    Args:
        snap_names (List[str]): Snaps to perform operations on.
        state (SnapState): State to set snap in.
        channel (str): Channel to track.
        classic (bool): Ensure snap is in classic confinement or not.
        cohort (str): Key of cohort to ensure snap belongs to (Default: "").

    Raises:
        SnapHandlerError: Raised if install or refresh operation fails.
    """
    snaps = {"success": [], "failed": []}

    for s in snap_names:
        try:
            snap = _Cache[s]
            if state is SnapState.ABSENT:
                snap.ensure(state=SnapState.ABSENT)
            else:
                snap.ensure(
                    state=state, classic=classic, channel=channel, cohort=cohort
                )
            snaps["success"].append(snap)
        except SnapHandlerError:
            snaps["failed"].append(s)

    if len(snaps["failed"]):
        raise SnapHandlerError(
            f"Failed to install or refresh snap(s): {', '.join(list(snaps['failed']))}"
        )

    return snaps["success"] if len(snaps["success"]) > 1 else snaps["success"][0]


def install_local(
    filename: str,
    classic: bool = False,
    devmode: bool = False,
    dangerous: bool = False,
) -> Snap:
    """Install snap package using local .snap file.

    Args:
        filename (str): The path to a local .snap file to install.
        classic (bool): Whether --classic should be passed to
            install snap with classic confinement (Default: False).
        dangerous (bool): Whether --dangerous should be passed to
            install snap without a signature (Default: False).
        devmode (bool): Whether --devmode should be passed to
            install snap with devmode confinement (Default: False).

    Raises:
        SnapHandlerError: Raised if local .snap package fails to install.
    """
    _cmd = [
        "snap",
        "install",
        filename,
    ]
    if classic:
        _cmd.append("--classic")
    if devmode:
        _cmd.append("--devmode")
    if dangerous:
        _cmd.append("--dangerous")
    try:
        result = subprocess.check_output(_cmd, universal_newlines=True).splitlines()[-1]
        snap_name, _ = result.split(" ", 1)
        snap_name = ansi_filter.sub("", snap_name)

        c = SnapCache()

        return c[snap_name]
    except CalledProcessError as e:
        raise SnapHandlerError(f"Could not install snap {filename}: {e.output}")


def alias(alias_snap: str, app: str, alias: str, wait: bool = True) -> None:
    """Add an alias for a snap command.

    Args:
        alias_snap (str): Snap that provides the app that an alias will be created for.
        app (str): App inside snap package to create alias for.
        alias (str): Alias to create.
        wait (bool):
            True - wait for alias operation to finish.
            False - do not wait for alias operation to finish (Default: True).

    Raises:
        SnapHandlerError: Raised if alias operation fails.
    """
    _cmd = ["snap", "alias", f"{alias_snap}.{app}", alias]
    if not wait:
        _cmd.append("--no-wait")

    try:
        subprocess.check_output(_cmd, universal_newlines=True)
    except subprocess.CalledProcessError:
        raise SnapHandlerError(
            f"Failed to create alias. Command used: {' '.join(_cmd)}"
        )


def unalias(alias_snap: str, wait: bool = True) -> None:
    """Remove an alias for a snap command.

    Args:
        alias_snap (str): Snap to remove aliases from..
        wait (bool):
            True - wait for unalias operation to finish.
            False - do not wait for unalias operation to finish (Default: True).

    Raises:
        SnapHandlerError: Raised if unalias operation fails.
    """
    _cmd = ["snap", "unalias", alias_snap]
    if not wait:
        _cmd.append("--no-wait")

    try:
        subprocess.check_output(_cmd, universal_newlines=True)
    except subprocess.CalledProcessError:
        raise SnapHandlerError(
            f"Failed to destroy alias. Command used: {' '.join(_cmd)}"
        )


def connect(
    plug: str,
    plug_snap: str,
    slot_snap: str = None,
    slot: str = None,
    wait: bool = True,
) -> None:
    """Connect a snap plug to a slot.

    Args:
        plug (str): Plug to connect to slot.
        plug_snap (str): Snap that provides plug.
        slot_snap (str): Snap that provides slot (Default: None).
        slot (str): Slot to connect to (Default: None).
        wait (bool):
            True - wait for connect operation to finish.
            False - do not wait for connect operation to finish (Default: True).

    Raises:
        SnapHandlerError: Raised if connect operation fails.
    """
    _cmd = ["snap", "connect", f"{plug_snap}:{plug}"]
    if slot_snap is not None and slot is not None:
        _cmd.append(f"{slot_snap}:{slot}")
    if slot_snap is not None and slot is None:
        _cmd.append(slot_snap)
    if slot_snap is None and slot is not None:
        _cmd.append(f":{slot}")
    if not wait:
        _cmd.append("--no-wait")

    try:
        subprocess.check_output(_cmd, universal_newlines=True)
    except subprocess.CalledProcessError:
        raise SnapHandlerError(f"Failed to connect. Command used: {' '.join(_cmd)}")


def disconnect(
    plug: str,
    plug_snap: str = None,
    slot_snap: str = None,
    slot: str = None,
    wait: bool = True,
    forget: bool = False,
) -> None:
    """Disconnect a snap plug from a slot.

    Args:
        plug (str): Plug to disconnect from slot.
        plug_snap (str): Snap that provides plug (Default: None).
        slot_snap (str): Snap that provides slot (Default: None).
        slot (str): Slot to disconnect from (Default: None).
        wait (bool):
            True - wait for disconnect operation to finish.
            False - do not wait for disconnect operation to finish (Default: True).
        forget (bool):
            True - forget remembered state about connection.
            False - do not forget remember state about connection (Default: False).

    Raises:
        SnapHandlerError: Raised if disconnect operation fails.
    """
    _cmd = ["snap", "disconnect", f":{plug}"]
    if plug_snap is not None:
        _cmd[2] = f"{plug_snap}:{plug}"
    if slot_snap is not None and slot is not None:
        _cmd.append(f"{slot_snap}:{slot}")
    if not wait:
        _cmd.append("--no-wait")
    if forget:
        _cmd.append("--forget")

    try:
        subprocess.check_output(_cmd, universal_newlines=True)
    except subprocess.CalledProcessError:
        raise SnapHandlerError(
            f"Failed to create alias. Command used: {' '.join(_cmd)}"
        )


def _system_set(config_item: str, value: str) -> None:
    """Helper for setting snap system config values.

    Args:
        config_item (str): Name of snap system setting. E.g. 'refresh.hold'.
        value (str): Value to assign.

    Raises:
        SnapHandlerError: Raised if system set operation fails.
    """
    _cmd = ["snap", "set", "system", f"{config_item}={value}"]
    try:
        subprocess.check_call(_cmd, universal_newlines=True)
    except CalledProcessError:
        raise SnapHandlerError(
            f"Failed setting system config '{config_item}' to '{value}'"
        )


def hold_refresh(days: int = 90) -> None:
    """Set the system-wide snap refresh hold.

    Args:
        days (int): Number of days to hold system refreshes for. Maximum 90.
            Set to zero to remove hold.

    Warnings:
        This method does not currently support snap's experimental `--hold` feature.
    """
    if not type(days) == int or days > 90:
        raise ValueError(f"Days must be an int between 1 and 90. Not {days}.")
    elif days == 0:
        _system_set("refresh.hold", "")
    else:
        # Add the number of days to current time
        target_date = datetime.now(timezone.utc).astimezone() + timedelta(days=days)
        # Format for the correct datetime format
        hold_date = target_date.strftime("%Y-%m-%dT%H:%M:%S%z")
        # Python dumps the offset in format '+0100', we need '+01:00'
        hold_date = "{0}:{1}".format(hold_date[:-2], hold_date[-2:])
        # Actually set the hold date
        _system_set("refresh.hold", hold_date)



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Abstract base class for injectable objects."""

import base64
import hashlib
import pickle
from abc import ABC, abstractmethod
from typing import Dict


class InjectionError(Exception):
    """Base error for classes that inherit from Injectable."""


class Injectable(ABC):
    """Abstract metaclass that provides core methods needed by all injectable objects."""

    @classmethod
    def _loads(cls, checksum: str, data: str) -> object:
        """Alternative constructor to load previously initialized object.

        Args:
            checksum (str): Checksum to verify authenticity of serialized object.
            data (str): Path to file containing serialized object.

        Returns:
            (object): Deserialized, verified object.
        """
        if type(data) != str:
            raise InjectionError(f"Cannot load object {data}. {type(data)} != str")

        _ = base64.b64decode(data)
        if checksum != hashlib.sha224(_).hexdigest():
            raise InjectionError("Hashes do not match. Will not load untrusted object.")

        _ = pickle.loads(_)
        posargs = [
            value for key, value in _.__dict__.items() if not key.startswith("_")
        ]
        hiddenargs = {
            key: value for key, value in _.__dict__.items() if key.startswith("_")
        }
        new_cls = cls(*posargs)
        [setattr(new_cls, key, value) for key, value in hiddenargs.items()]
        return new_cls

    def _dumps(self, **kwargs) -> Dict[str, str]:
        """Prepare object for injection.

        Returns:
            (Dict[str, str]):
                checksum (str): Checksum to verify authenticity of serialized object.
                data (str): Base64 encoded string containing serialized object.
                injectable (str): Injectable to run inside remote environment.
        """
        pickle_data = pickle.dumps(self)
        checksum = hashlib.sha224(pickle_data).hexdigest()
        data = base64.b64encode(pickle_data).decode()
        return {
            "checksum": checksum,
            "data": data,
            "injectable": self._injectable(
                {"checksum": checksum, "data": data}, **kwargs
            ),
        }

    @abstractmethod
    def _injectable(self, data: Dict[str, str], **kwargs) -> str:
        """Python script to run inside of test environment provider."""



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Abstractions for uploading and downloading files from test environments."""

import copy
import os
import pathlib
import shutil
import tarfile
import tempfile
import textwrap
from io import BytesIO, StringIO
from typing import Dict, Union

from cleantest.meta import Injectable


class FileError(Exception):
    """Base error for File class."""


class InjectableModeError(Exception):
    """Raised when an invalid injection mode has been passed to `_injectable`."""


class File(Injectable):
    """Represents a file that can be shared between host and test environment.

    Args:
        src (Union[str, pathlib.Path, StringIO, BytesIO]):
            Where to load file from.
        dest (Union[str, pathlib.Path]): Where to dump file to.
        overwrite (bool):
            True - overwrite file if it already exists when dumping.
            False - raise error if file already exists when dumping.
    """

    def __init__(
        self,
        src: Union[str, pathlib.Path, bytes, StringIO, BytesIO],
        dest: Union[str, pathlib.Path],
        overwrite: bool = False,
    ) -> None:
        self.src = src
        self.dest = pathlib.Path(dest)
        self.overwrite = overwrite
        self.__data = None

    def load(self) -> None:
        """Load file from specified source.

        Raises:
            FileNotFoundError: Raised if file is not found.
            FileError: Raised if source is a directory rather than a file.
        """
        with tempfile.TemporaryDirectory() as tmp_dir:
            _ = pathlib.Path(tmp_dir)
            with tarfile.open(pathlib.Path(tmp_dir) / "_", "w:gz") as tar:
                old_pwd = os.getcwd()
                os.chdir(_)
                if type(self.src) == str or isinstance(self.src, pathlib.Path):
                    data = pathlib.Path(self.src)
                    if data.exists() is False:
                        raise FileNotFoundError(f"Could not find {self.src}.")

                    if data.is_dir():
                        raise FileError(
                            f"{self.src} is a directory. Use Dir class instead."
                        )
                    shutil.copy(data, _ / "data")
                    tar.add("data")
                elif isinstance(self.src, StringIO) or isinstance(self.src, BytesIO):
                    data = copy.deepcopy(self.src)
                    placeholder = pathlib.Path("data")
                    placeholder.write_text(data.read()) if isinstance(
                        self.src, StringIO
                    ) else placeholder.write_bytes(data.read())
                    tar.add(placeholder)
                else:
                    raise FileError(
                        (
                            "Expected type str, os.PathLike, StringIO, or BytesIO, "
                            f"not {type(self.src)}."
                        )
                    )

                os.chdir(old_pwd)

            self.__data = pathlib.Path(tar.name).read_bytes()

    def dump(self) -> None:
        """Dump directory to specified destination.

        Raises:
            FileExistsError: Raised if directory already exists and overwrite=False.
            FileError: Raised if no data has been loaded prior to calling dump().
        """
        if self.dest.exists() and self.overwrite is False:
            raise FileExistsError(
                (
                    f"{self.dest} already exists. Set overwrite = True to "
                    f"overwrite {self.dest}."
                )
            )

        if self.__data is None:
            raise FileError("Nothing to write.")

        with tempfile.TemporaryDirectory() as tmp_dir, tarfile.open(
            fileobj=BytesIO(self.__data), mode="r:gz"
        ) as tar:
            tar.extractall(tmp_dir)
            shutil.copy((pathlib.Path(tmp_dir) / "data"), self.dest)

    def _injectable(self, data: Dict[str, str], **kwargs) -> str:
        """Generate injectable script that will be run inside the test environment.

        Args:
            data (Dict[str, str]): Data that needs to be in injectable script.
                - checksum (str): SHA224 checksum to verify authenticity of object.
                - data (str): Base64 encoded object to inject.
            **kwargs:
                mode (str): "push" or "pull" object to/from test environment instance.

        Raises:
            InjectableModeError: Raised if invalid mode has been passed.

        Returns:
            (str): Injectable script.
        """
        _ = kwargs.get("mode", None)
        if _ not in {"push", "pull"}:
            InjectableModeError(
                f"Invalid mode: {_}. Please set mode to either 'push' or 'pull'."
            )
        elif _ == "push":
            return textwrap.dedent(
                f"""
                #!/usr/bin/env python3
                
                from {self.__module__} import {self.__class__.__name__}
                
                _ = {self.__class__.__name__}._loads("{data['checksum']}", "{data['data']}")
                _.dump()
                """
            ).strip("\n")
        else:
            return textwrap.dedent(
                f"""
                #!/usr/bin/env python3
                import json
                import sys
                
                from {self.__module__} import {self.__class__.__name__}
                
                _ = {self.__class__.__name__}._loads("{data['checksum']}", "{data['data']}")
                _.load()
                print(json.dumps(_._dumps()), file=sys.stdout)
                """
            ).strip("\n")

    def __repr__(self) -> str:
        """String representation of File."""
        attrs = ", ".join(
            f"{k}={v}" for k, v in self.__dict__.items() if not k.startswith("_")
        )
        return f"{self.__class__.__name__}({attrs})"



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Environment variable management for test providers."""

import os
from typing import Any, Dict, Optional

from cleantest.meta.mixins import Resettable


class Env(Resettable):
    """Manage environment data for test environments."""

    _env = {}

    def __new__(cls) -> "Env":
        """Create new Env object instance.

        Returns:
            (Env): New object instance.
        """
        if not hasattr(cls, "_instance"):
            cls._instance = super(Env, cls).__new__(cls)
        return cls._instance

    def reset(self) -> None:
        """Reset environment information."""
        self._env = {}

    def add(self, env_mapping: Dict[str, Any]) -> None:
        """Add new values to environment.

        Args:
            env_mapping (Dict[str, Any]):
                Key, value mapping to add to the environment store.
        """
        self._env.update(env_mapping)

    def remove(self, env_var: str) -> None:
        """Remove an environment variable from store.

        Does nothing if environment variable does not exist in store.

        Args:
            env_var (str): Environment variable to Remove from the store.
        """
        if env_var in self._env.keys():
            del self._env[env_var]

    def get(self, env_var: str) -> Optional[Any]:
        """Retrieve environment variable from store.

        Args:
            env_var (str): Environment variable to retrieve.

        Returns:
            (Optional[Any]): Environment variable value.
                Returns None if variable is not in store.
        """
        try:
            if type(self._env[env_var]) == list:
                return os.pathsep.join(self._env[env_var])
        except KeyError:
            return None

    def dump(self) -> Dict[str, Any]:
        """Dump environment store as a dictionary.

        Returns:
            (Dict[str, Any]): Environment store as a dictionary.
        """
        result = {}
        for k, v in self._env.items():
            if type(v) == list:
                result.update({k: os.pathsep.join(v)})
            else:
                result.update({k: v})

        return result



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Abstractions for uploading and downloading directories from test environments."""

import os
import pathlib
import tarfile
import tempfile
from io import BytesIO
from typing import Iterable, Union

from cleantest.data.file import File


class DirectoryError(Exception):
    """Base error for Dir class."""


class DirectoryExistsError(Exception):
    """Raised when a directory already exists on the host."""


class DirectoryNotFoundError(Exception):
    """Raised when a directory is not found on the host."""


def _strip_tar(
    tar: tarfile.TarFile, n_components: int = 1
) -> Iterable[tarfile.TarInfo]:
    """Strip components from a tarfile to a specified level.

    Args:
        tar (tarfile.TarFile): Opened tar archive to strip.
        n_components (int): Number of levels to strip from tar file (Default: 1).

    Yields:
        (Iterable[tarfile.TarInfo]): TarInfo object with its path property modified.
    """
    for member in tar.getmembers():
        member_path = pathlib.Path(member.path)
        member.path = member_path.relative_to(*member_path.parts[:n_components])
        yield member


class Dir(File):
    """Represents a directory that can be shared between host and test environment.

    Args:
        src (Union[str, os.PathLike]): Where to load directory from.
        dest (Union[str, os.PathLike]): Where to dump directory to.
        overwrite (bool):
            True - overwrite directory if it already exists when dumping.
            False - raise error if directory already exists when dumping.
    """

    def __init__(
        self,
        src: Union[str, os.PathLike],
        dest: Union[str, os.PathLike],
        overwrite: bool = False,
    ) -> None:
        super().__init__(pathlib.Path(src), dest, overwrite)

    def load(self) -> None:
        """Load directory from specified source.

        Raises:
            DirectoryNotFoundError: Raised if directory is not found.
            NotADirectoryError: Raised if source is a file rather than a directory.
        """
        if self.src.exists() is False:
            raise DirectoryNotFoundError(f"Could not find {self.src}.")

        if self.src.is_file():
            raise NotADirectoryError(f"{self.src} is a file. Use File class instead.")

        _ = os.getcwd()
        os.chdir(os.sep.join(str(self.src).split(os.sep)[:-1]))
        with tempfile.NamedTemporaryFile() as fin:
            with tarfile.open(fin.name, "w:gz") as tar:
                tar.add(self.src.name)
            self.__data = pathlib.Path(fin.name).read_bytes()
        os.chdir(_)

    def dump(self) -> None:
        """Dump directory to specified destination.

        Raises:
            DirectoryExistsError: Raised if directory already exists and overwrite=False.
            DirectoryError: Raised if no data has been loaded prior to calling dump().
        """
        if self.dest.exists() and self.overwrite is False:
            raise DirectoryExistsError(
                f"{self.dest} already exists. Set overwrite = True to overwrite {self.dest}."
            )

        if self.__data is None:
            DirectoryError("Nothing to write.")

        with tarfile.open(fileobj=BytesIO(self.__data), mode="r:gz") as tar:
            tar.extractall(self.dest, members=_strip_tar(tar))



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Handler for installing and managing Debian packages inside test environment instances."""

# Note: Support for third-party repositories is not implemented yet.

import os
import pathlib
import re
import subprocess
from shutil import which
from typing import Iterable, List, Optional, Union

from cleantest.meta import Result
from cleantest.meta.utils import detect_os_variant


class AptHandlerError(Exception):
    """Raised when apt handler encounters any errors."""


def _is_apt_available() -> None:
    """Check if apt is available on test environment instance.

    Raises:
        AptHandlerError: Raised if apt is not available.
    """
    if which("apt") is None:
        raise AptHandlerError(
            (
                f"apt is not supported on {detect_os_variant()}. "
                f"`from cleantest.utils import run` can be used instead."
            )
        )


def _apt(
    command: str, packages: Union[str, List[str]], optargs: Optional[List[str]] = None
) -> Result:
    """Wrap package management commands from Debian/Ubuntu commands.

    Args:
        command (str): Command to execute.
        packages (Union[str, List[str]]): Names of packages to operate on.
        optargs (Optional[List[str]]): Optional arguments to pass to apt.

    Raises:
        AptHandlerError: Raised if executed command fails.

    Returns:
        (Result): Captured exit code, stdout, and stderr.
    """
    packages = [packages] if type(packages) == str else packages
    optargs = optargs if optargs is not None else []
    _cmd = ["apt", "-y", *optargs, command, *packages]
    try:
        process = subprocess.run(
            _cmd,
            env={"DEBIAN_FRONTEND": "noninteractive", "PATH": os.getenv("PATH")},
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        return Result(process.returncode, process.stdout, process.stderr)

    except subprocess.CalledProcessError:
        raise AptHandlerError(
            f"Could not perform command {_cmd} on the following packages: {', '.join(packages)}"
        )


def update() -> None:
    """Update apt cache on the test environment instance.

    Raises:
        AptHandlerError: Raised if unable to update apt cache.
    """
    _is_apt_available()
    try:
        subprocess.check_call(
            ["apt", "-y", "update"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError as e:
        raise AptHandlerError(f"Failed to update apt cache. Reason: {e}.")


def install(*packages: str, version: Optional[str] = None) -> None:
    """Install Debian package on test environment instance.

    Args:
        *packages (str): Names of packages to install.
        version (Optional[str]): Version of package to install.
            Only set if installing one package (Default: None).

    Raises:
        AptHandlerError: Raised if an error is encountered when installing packages.
    """
    _is_apt_available()
    if len(packages) == 0:
        raise AptHandlerError("No package names passed.")
    if len(packages) > 1 and version:
        raise AptHandlerError("Version should not be set if packages to install > 1.")

    # Note: Use --force-confold to keep old configuration file rather than generating a new one.
    if version:
        _apt(
            "install",
            f"{packages[0]}={version}",
            optargs=["--option=Dpkg::Options::=--force-confold"],
        )
    else:
        _apt(
            "install", [*packages], optargs=["--option=Dpkg::Options::=--force-confold"]
        )


def install_local(*package_paths: str) -> None:
    """Install local Debian package on test environment instance."""
    _is_apt_available()
    if len(package_paths) == 0:
        raise AptHandlerError("No file paths to packages passed.")
    for path in package_paths:
        if not pathlib.Path(path).exists():
            raise FileNotFoundError(f"Could not locate deb archive {path}.")
        _apt(
            "install", path, optargs=["--option=Dpkg::Options::=--force-confold", "-f"]
        )


def remove(*packages: str) -> None:
    """Remove package from test environment instance.

    Args:
        *packages (str): Names of packages to remove.

    Raises:
        AptHandlerError: Raised if an error is encountered when removing packages.
    """
    _is_apt_available()
    if len(packages) == 0:
        raise AptHandlerError("No package names passed.")
    _apt("remove", [*packages])


def purge(*packages: str) -> None:
    """Purge package from test environment instance.

    Args:
        *packages (str): Names of packages to purge.
    """
    _is_apt_available()
    if len(*packages) == 0:
        raise AptHandlerError("No package names passed.")
    _apt("purge", [*packages])


def installed(*packages: str) -> Iterable[bool]:
    """Check if packages are installed inside test environment instance.

    Args:
        *packages (str): Names of packages to check if they are installed.

    Yields:
        (Iterable[bool]): Status of package. True: Installed - False: Not Installed.
    """
    _is_apt_available()
    if len(*packages) == 0:
        raise AptHandlerError("No package names passed.")
    check_installed = re.compile(r"\[installed]")
    for package in packages:
        yield True if check_installed.match(_apt("list", package).stdout) else False



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Manager for installing charm libraries inside remote processes."""

import json
import pathlib
import subprocess
import sys
import textwrap
from shutil import which
from typing import Dict, List, Union

from cleantest.meta import BasePackage, BasePackageError
from cleantest.meta.mixins import SnapdSupport
from cleantest.utils import snap


class CharmlibPackageError(BasePackageError):
    """Base error for Charmlib package handler."""


class Charmlib(BasePackage, SnapdSupport):
    """Represents `charmcraft fetch-lib`. See: https://juju.is/docs/sdk/charm-libraries.

    Args:
        auth_token_path (pathlib.Path): Path to authentication token for Charmhub.
        charmlibs (List[str]): List of charmlibs to install.
    """

    def __init__(
        self,
        auth_token_path: str,
        charmlibs: Union[str, List[str]],
    ) -> None:
        self.auth_token_path = pathlib.Path(auth_token_path)
        self.charmlibs = [charmlibs] if type(charmlibs) == str else charmlibs
        self._auth_token = None

        if auth_token_path is None:
            raise CharmlibPackageError(
                "No authentication token for Charmhub specified."
            )

        if charmlibs is None:
            raise CharmlibPackageError("No charm libraries specified.")

    def _run(self) -> None:
        """Run Charmlib package handler."""
        self._setup()
        self._handle_charm_lib_install()
        print(json.dumps({"PYTHONPATH": "/root/lib"}), file=sys.stdout)

    def _setup(self) -> None:
        """Set up and install dependencies needed for charm libraries."""
        self._install_snapd()
        if which("charmcraft") is None:
            snap.install("charmcraft", classic=True)

    def _handle_charm_lib_install(self) -> None:
        """Install charm libraries inside test environment.

        Raises:
            CharmlibPackageError: Raised if charm library fails to install.
        """
        for charm in self.charmlibs:
            cmd = ["/snap/bin/charmcraft", "fetch-lib", charm]
            try:
                subprocess.run(
                    cmd,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    check=True,
                    env={"CHARMCRAFT_AUTH": self._auth_token},
                    cwd="/root",
                )
            except subprocess.CalledProcessError:
                raise CharmlibPackageError(
                    (
                        f"Failed to install charm library {charm} "
                        f"using the following command: {' '.join(cmd)}"
                    )
                )

    def _dumps(self) -> Dict[str, str]:
        """Prepare Charmlib object for injection.

        Raises:
            FileNotFoundError: Raised if authentication token is not found.

        Returns:
            (Dict[str, str]):
                checksum (str): Checksum to verify authenticity of serialized Charmlib object.
                data (str): Base64 encoded string containing serialized Charmlib object.
                injectable (str): Injectable to run inside remote environment.
        """
        if not self.auth_token_path.exists() or not self.auth_token_path.is_file():
            raise FileNotFoundError(
                f"Could not find authentication token {self.auth_token_path}"
            )

        self._auth_token = self.auth_token_path.read_text()

        return super()._dumps()

    def _injectable(self, data: Dict[str, str], **kwargs) -> str:
        """Generate injectable script that will be used to install charm libraries.

        Args:
            data (Dict[str, str]): Data that needs to be in injectable script.
                - checksum (str): SHA224 checksum to verify authenticity of Charmlib object.
                - data (str): Base64 encoded Charmlib object to inject.

        Returns:
            (str): Injectable script.
        """
        return textwrap.dedent(
            f"""
            #!/usr/bin/env python3
            
            from {self.__module__} import {self.__class__.__name__}
            
            holder = {self.__class__.__name__}._loads("{data['checksum']}", "{data['data']}")
            holder._run()
            """
        ).strip("\n")



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Detect thread count on host system."""

import os


def thread_count() -> int:
    """Get number of allowable threads on host system.

    Returns:
        (int): Number of allowable threads (Default: os.cpu_count())
    """
    env_var = os.getenv("CLEANTEST_NUM_THREADS")
    return env_var if env_var is not None and type(env_var) == int else os.cpu_count()



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Public utilities used inside cleantest.

You can use these utilities for developing third-party plugins.
"""

from .detect_os import detect_os_variant
from .thread_count import thread_count



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Public mixins used inside cleantest."""

from .dict_like import DictLike
from .enhanced_enum import EnhancedEnum
from .resettable import Resettable
from .snapd_support import SnapdSupport



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Run commands inside test environment instances."""

import shlex
import subprocess
from typing import Any, Dict, Iterable, Optional

from cleantest.meta import Result


class CommandExecutionError(Exception):
    """Raised if run fails to execute a command."""


def run(
    *commands: str,
    env: Optional[Dict[str, Any]] = None,
    cwd: Optional[str] = None,
) -> Iterable[Result]:
    """Run a command inside a test environment instance.

    Args:
        *commands (str): Commands to execute.
        env (Optional[Dict[str, Any]]): Environment to pass to commands (Default: None).
        cwd (Optional[str]): Directory to execute commands inside (Default: None).

    Yields:
        (Iterable[Result]): Captured results of executed commands.
    """
    if len(commands) == 0:
        raise CommandExecutionError("No commands passed.")

    for command in commands:
        try:
            res = subprocess.run(
                shlex.split(command),
                env=(env if env else None),
                cwd=(cwd if cwd else None),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise CommandExecutionError(
                f"Failed to execute command {command}. Reason: {e}."
            )

        yield Result(res.returncode, res.stdout, res.stderr)



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Private metaclass that provides tooling needed by configurer classes."""

import copy
from collections import deque
from typing import Deque, Union

from cleantest.control.hooks import StartEnvHook, StopEnvHook
from cleantest.meta.mixins import Resettable


class BaseConfigurerError(Exception):
    """Base error for all configurer classes."""


class DuplicateHookNameError(Exception):
    """Raised when more than one hook of the same type have the same name."""


class _HookRegistry(Resettable):
    """Centrally store hooks for use by test environment providers."""

    metadata = set()
    startenv = deque()
    stopenv = deque()

    def __new__(cls) -> "_HookRegistry":
        if not hasattr(cls, "_instance"):
            cls._instance = super(_HookRegistry, cls).__new__(cls)
        return cls._instance

    def reset(self) -> None:
        """Reset metadata and hook queues."""
        self.metadata = set()
        self.startenv = deque()
        self.stopenv = deque()

    def lint(self, hook: Union[StartEnvHook, StopEnvHook]) -> None:
        """Lint hooks to ensure they compliant with set restrictions."""
        lint = [
            h
            for h in self.metadata
            if hook.name == h[0] and hook.__class__.__name__ == h[1]
        ]
        if len(lint) > 0:
            raise DuplicateHookNameError(
                (
                    f"Hook type {hook.__class__.__name__} with name "
                    f"{hook.name} already exists."
                )
            )


class BaseConfigurer(Resettable):
    """Base configure mixin for configurers."""

    def reset(self) -> None:
        """Reset the hook registry."""
        _HookRegistry().reset()

    def register_hook(self, *hooks: Union[StartEnvHook, StopEnvHook]) -> None:
        """Register hooks in the hook registry.

        Args:
            *hooks (Union[StartEnvHook, StopEnvHook]): Hooks to register.
        """
        dispatch = {
            StartEnvHook.__name__: _HookRegistry().startenv,
            StopEnvHook.__name__: _HookRegistry().stopenv,
        }
        for hook in hooks:
            _HookRegistry().lint(hook)
            dispatch[hook.__class__.__name__].appendleft(hook)
            _HookRegistry().metadata.add((hook.name, hook.__class__.__name__))

    def unregister_hook(self, *hook_names: str) -> None:
        """Unregister hooks from the hook registry.

        Args:
            *hook_names (str): Names of hooks to unregister.
        """
        dispatch = {
            StartEnvHook.__name__: _HookRegistry().startenv,
            StopEnvHook.__name__: _HookRegistry().stopenv,
        }
        for name in hook_names:
            for hook in _HookRegistry().metadata:
                if name == hook[0]:
                    [
                        dispatch[hook[1]].remove(i)
                        for i in dispatch[hook[1]]
                        if i.name == name
                    ]
                    _HookRegistry().metadata.remove(hook)

    @property
    def startenv_hooks(self) -> Deque[StartEnvHook]:
        """Retrieve hooks that will run when the test environment starts.

        Returns:
            (Deque[StartEnvHook]): Deque containing start environment hooks.
        """
        return copy.deepcopy(_HookRegistry().startenv)

    @property
    def stopenv_hooks(self) -> Deque[StopEnvHook]:
        """Retrieve hooks that will run when the test environment stops.

        Returns:
            (Deque[StopEnvHook]): Deque containing stop environment hooks.
        """
        return copy.deepcopy(_HookRegistry().stopenv)



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Abstract class for test environment instance handlers."""

import re
import tempfile
from abc import ABC, abstractmethod
from typing import Any, Dict, List

from .result import Result


class BaseEntrypointError(Exception):
    """Base error for test run entrypoints."""


class BaseHarnessError(Exception):
    """Base error for test environment provider harnesses."""


class BaseEntrypoint(ABC):
    """Metaclass for test environment provider entrypoints.

    Entrypoints define the tooling stubs needed to start running tests.
    """

    @abstractmethod
    def run(self) -> Dict[str, Result]:
        """Run handler for test environment."""


class BaseHarness(ABC):
    """Base class for test environment provider harnesses.

    Harnesses wrap around testlets to ensure they are run inside the
    specified test environment instance.
    """

    @abstractmethod
    def _exists(self) -> bool:
        """Check if test environment already exists."""

    @abstractmethod
    def _init(self) -> None:
        """Build and initialize test environment instance."""

    @abstractmethod
    def _execute(self) -> Result:
        """Execute testlet inside test environment instance."""

    @abstractmethod
    def _teardown(self) -> None:
        """Destroy test environments if not told to preserve."""

    @abstractmethod
    def _handle_start_env_hooks(self) -> None:
        """Handle StartEnv hooks."""

    @abstractmethod
    def _handle_stop_env_hooks(self) -> None:
        """Handle StopEnv hooks."""

    def _make_testlet(self, src: str, name: str, remove: List[Any] = None) -> str:
        """Construct Python source file to be run in subroutine.

        Args:
            src (str): Source code of testlet.
            name (str): Name of testlet.
            remove (List[Any]): String patterns to remove from source code in regex form.

        Returns:
            (str): Injectable testlet.

        Future:
            This will need more advanced logic if tests accept arguments.
        """
        if remove is not None:
            for pattern in remove:
                src = re.sub(pattern, "", src)

        with tempfile.TemporaryFile(mode="w+t") as _:
            content = [
                "#!/usr/bin/env python3\n",
                f"{src}\n",
                f"{name}()\n",
            ]
            _.writelines(content)
            _.seek(0)
            testlet = _.read()

        return testlet



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Mixin for enhancing Enum objects."""

from enum import Enum
from typing import Any, List, Tuple


class EnhancedEnum(Enum):
    """Mixin for giving enums extra methods for convenience."""

    @classmethod
    def items(cls) -> List[Tuple[str, Any]]:
        """Returns items of an Enum."""
        return [(c.name, c.value) for c in cls]



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""LXD test environment provider functions and utilities."""

import functools
from typing import Callable, Iterable, List, Optional, Tuple, Union

from cleantest.control import Configure, Env
from cleantest.meta import Result
from cleantest.meta.utils import thread_count

from ._lxd_harness import LXDProviderEntrypoint


class lxd:  # noqa N801
    """LXD test environment provider.

    Args:
        name (str): Name for test environment (Default: "test").
        image (List[str]):
            LXD image to use for test environment (Default: ["ubuntu-jammy-amd64"]).
        preserve (bool): Preserve test environment after test has completed (Default: True).
        parallel (bool): Run test environment instances in parallel (Default: False).
        num_threads (int): Number of threads to use when running
            test environment instances in parallel (Default: None).
    """

    def __init__(
        self,
        name: str = "test",
        image: Union[str, List[str]] = ["ubuntu-jammy-amd64"],
        preserve: bool = True,
        parallel: bool = False,
        num_threads: Optional[int] = None,
    ) -> None:
        self._name = name
        self._image = [image] if type(image) == str else image
        self._preserve = preserve
        self._env = Env()
        self._parallel = parallel
        self._lxd_config = Configure("lxd")

        if (type(num_threads) != int or num_threads < 1) and self._parallel is True:
            self._num_threads = thread_count()
        elif type(num_threads) == int and self._parallel is True:
            self._num_threads = num_threads

    def __call__(self, func: Callable) -> Callable:
        """Callable for lxd decorator."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Iterable[Tuple[str, Result]]:
            handler = (
                LXDProviderEntrypoint(strategy="parallel", func=func, **self.__dict__)
                if self._parallel is True
                else LXDProviderEntrypoint(
                    strategy="serial", func=func, **self.__dict__
                )
            )
            return handler.run()

        return wrapper

    @classmethod
    def target(
        cls,
        *instances: str,
        parallel: bool = False,
        num_threads: Optional[int] = None,
    ) -> Callable:
        """Target specific LXD test environment instances by name.

        Args:
            *instances (str): Test environment instance name.
            parallel (bool):
                Run test environment instances in parallel (Default: False).
            num_threads (int): Number of threads to use when running
                test environment instances in parallel (Default: None).

        Returns:
            (Callable): Wrapped function.
        """

        def decorator(func: Callable) -> Callable:
            @functools.wraps(func)
            def wrapper(*args, **kwargs) -> Iterable[Tuple[str, Result]]:
                if (type(num_threads) != int or num_threads < 1) and parallel is True:
                    _num_threads = num_threads
                elif type(num_threads) == int and parallel is True:
                    _num_threads = num_threads
                else:
                    _num_threads = None
                _ = {
                    "_target_instances": [*instances],
                    "_env": Env(),
                    "_lxd_config": Configure("lxd"),
                    "_num_threads": _num_threads,
                }
                handler = (
                    LXDProviderEntrypoint(strategy="parallel_target", func=func, **_)
                    if parallel is True
                    else LXDProviderEntrypoint(strategy="serial_target", func=func, **_)
                )
                return handler.run()

            return wrapper

        return decorator



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Mixin for objects that need to emulate a dictionary."""

from typing import Any, Dict, Iterable, Tuple


class DictLike:
    """Mixin for objects that need to emulate a dictionary."""

    def dict(self) -> Dict:
        """Return class as a unidirectional dictionary.

        Returns:
            (Dict): Class as a dictionary.
        """
        return self.__process_dict(self.__dict__)

    def keys(self, all_keys: bool = False) -> Iterable[Any]:
        """Get dictionary keys.

        Args:
            all_keys (bool):
                If False, only get top-level dictionary keys.
                If True, get all keys in the multi-node dictionary.
                (Default: False).

        Returns:
            (Iterable[Any]): Iterable containing dictionary keys.
        """
        if not all_keys:
            return iter(k for k in self.__dict__.keys())
        else:
            return self.__process_keys(self.dict())

    def values(self, all_values: bool = False) -> Iterable[Any]:
        """Get dictionary values.

        Args:
            all_values (bool):
                If False, only get top-level dictionary values.
                If True, get all values in the multi-node dictionary.
                (Default: False).

        Returns:
            (Iterable[Any]): Iterable containing dictionary values.
        """
        if not all_values:
            return iter(v for v in self.__dict__.values())
        else:
            return self.__process_values(self.dict())

    def items(self, all_items: bool = False) -> Iterable[Tuple[Any, Any]]:
        """Get dictionary items.

        Args:
            all_items (bool):
                If False, only get top-level dictionary items.
                If True, get all items in the multi-node dictionary.
                (Default: False).

        Returns:
            (Iterable[Tuple[Any, Any]]): Iterable containing dictionary items.
        """
        if not all_items:
            return iter(
                (k, v) for k, v in zip(self.__dict__.keys(), self.__dict__.values())
            )
        else:
            return iter(
                (k, v) for k, v in zip(self.keys(all_items), self.values(all_items))
            )

    def __process_dict(self, input_dict: Dict[Any, Any]) -> Dict[Any, Any]:
        """Process dictionary containing object with __dict__ attribute recursively.

        Args:
            input_dict (Dict): Dictionary containing object with __dict__ attribute.

        Returns:
            (Dict): Dictionary with containing objects converted to dictionaries.
        """
        result = {}
        for key, value in input_dict.items():
            if hasattr(value, "__dict__"):
                result.update({key: self.__process_dict(value.__dict__)})
            else:
                result.update({key: value})

        return result

    def __process_keys(self, input_dict: Dict[Any, Any]) -> Iterable[Any]:
        """Get all keys in a multi-node dictionary recursively.

        Args:
            input_dict (Dict[Any, Any]):
                Dictionary to get all keys from.

        Returns:
            (Iterable[Any]): All dictionary keys.
        """
        result = []
        for k, v in input_dict.items():
            if type(v) == dict:
                result.extend(self.__process_keys(v))
            else:
                result.append(k)

        return iter(result)

    def __process_values(self, input_dict: Dict[Any, Any]) -> Iterable[Any]:
        """Get all values in a multi-node dictionary recursively.

        Args:
            input_dict (Dict[Any, Any]):
                Dictionary to get all values from.

        Returns:
            (Iterable[Any]): All dictionary values.
        """
        result = []
        for k, v in input_dict.items():
            if type(v) == dict:
                result.extend(self.__process_values(v))
            else:
                result.append(v)

        return iter(result)



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Manager for installing snap packages inside remote processes."""

import pathlib
import textwrap
from enum import Enum
from typing import Dict, List, Union

from cleantest.meta import BasePackage, BasePackageError
from cleantest.meta.mixins import SnapdSupport
from cleantest.utils import snap


class SnapPackageError(BasePackageError):
    """Base error for Snap package handler."""


class Confinement(Enum):
    """Confinement modes for snap packages."""

    STRICT = "strict"
    CLASSIC = "classic"
    DEVMODE = "devmode"


class Plug:
    """Represents a snap plug.

    Args:
        snap (str): Name of snap that provides plug.
        name (str): Name of plug.
    """

    def __init__(self, snap: str, name: str) -> None:
        self.snap = snap
        self.name = name


class Slot:
    """Represents a snap slot.

    Args:
        snap (str): Name of snap that provides slot.
        name (str): Name of slot.
    """

    def __init__(self, snap: str = None, name: str = None) -> None:
        self.snap = snap
        self.name = name


class Connection:
    """Represents `snap connect`.

    Args:
        plug (Plug): Plug to connect.
        slot (Slot): Slot to connect to (Default: None).
        wait (bool): Wait for `snap connect` operation to complete (Default: True).

    Raises:
        SnapPackageError: Raised if lint rule fails.
    """

    def __init__(self, plug: Plug, slot: Slot = None, wait: bool = True) -> None:
        self._plug = plug
        self._slot = slot
        self._wait = wait
        self._lint()

    def _lint(self) -> None:
        """Lint inputs passed to class constructor."""
        if self._plug.snap is None or self._plug.name is None:
            raise SnapPackageError(
                f"Invalid plug: {self._plug.__dict__}. "
                "Plug must have an associated snap and name."
            )
        if (
            self._slot is not None
            and self._slot.snap is None
            and self._slot.name is None
        ):
            raise SnapPackageError(
                f"Invalid slot: {self._slot.__dict__}. "
                "Slot must at least have an associated snap or name."
            )

    def connect(self) -> None:
        """Perform `snap connect` operation."""
        snap.connect(
            self._plug.snap,
            self._plug.name,
            self._slot.snap,
            self._slot.name,
            wait=self._wait,
        )


class Alias:
    """Represents `snap alias`.

    Args:
        snap_name (str): Name of snap that provides the app to alias.
        app_name (str): Name of the app to alias.
        alias_name (str): Name of alias to create.
        wait (bool): Wait for `snap alias` operation to complete (Default: True).

    Raises:
        SnapPackageError: Raised if lint rule fails.
    """

    def __init__(
        self, snap_name: str, app_name: str, alias_name: str, wait: bool = True
    ) -> None:
        self._snap_name = snap_name
        self._app_name = app_name
        self._alias_name = alias_name
        self._wait = wait
        self._lint()

    def _lint(self) -> None:
        """Lint inputs passed to class constructor."""
        if (
            self._snap_name is None
            or self._app_name is None
            or self._alias_name is None
        ):
            holder = ", ".join(
                [
                    f"{key} = {value}"
                    for key, value in self.__dict__.items()
                    if value is None
                ]
            )
            raise SnapPackageError(f"Invalid alias: {holder} cannot be None.")

    def alias(self) -> None:
        """Perform `snap alias` operation."""
        snap.alias(self._snap_name, self._app_name, self._alias_name, self._wait)


class Snap(BasePackage, SnapdSupport):
    """Represents `snap install`.

    Args:
        snaps (List[str]): List of snaps to install (Default: None).
        local_snaps (List[str]): List of file paths to local snaps to install (Default: None).
        confinement (Confinement): Confinement mode to install snaps in
            (Default: Confinement.STRICT).
        channel (str): Channel to install snaps from (Default: None).
        cohort (str): Key of cohort that snap belongs to (Default: None).
        dangerous (bool): Install unsigned snaps (Default: False).
        connections (List[Connection]): List of connections to set up after snap is installed
            (Default: None).
        aliases (List[Alias]): List aliases to create after snap is installed (Default: None).

    Raises:
        SnapPackageError: Raised if class creation fails.
    """

    def __init__(
        self,
        snaps: Union[str, List[str]] = None,
        local_snaps: Union[str, List[str]] = None,
        confinement: Confinement = Confinement.STRICT,
        channel: str = None,
        cohort: str = None,
        dangerous: bool = False,
        connections: List[Connection] = None,
        aliases: List[Alias] = None,
    ) -> None:
        self.snaps = [snaps] if type(snaps) == str else snaps
        self.local_snaps = [local_snaps] if type(local_snaps) == str else local_snaps
        self._cached_local_snaps = set()
        self.confinement = confinement
        self.channel = channel
        self.cohort = cohort
        self.dangerous = dangerous
        self.connections = connections
        self.aliases = aliases

        if snaps is None and local_snaps is None:
            raise SnapPackageError("No snaps specified.")

        if not hasattr(Confinement, confinement.name):
            raise SnapPackageError(
                f"Invalid confinement {confinement.name}. "
                f"Must be either {', '.join([i.name for i in Confinement])}"
            )

    def _run(self) -> None:
        """Run Snap package handler."""
        self._setup()
        self._handle_snap_install()

    def _setup(self) -> None:
        """Install snapd inside test environment."""
        self._install_snapd()

    def _handle_snap_install(self) -> None:
        """Install snap packages inside test environment."""
        if self.snaps is not None:
            snap.install(
                self.snaps,
                channel=self.channel,
                classic=True if self.confinement == Confinement.CLASSIC else False,
                cohort=self.cohort if self.cohort is not None else "",
            )

        for pkg in self._cached_local_snaps:
            path = pathlib.Path.home().joinpath("tmp.snap")
            path.write_bytes(pkg)
            snap.install_local(
                str(path),
                classic=True if self.confinement == Confinement.CLASSIC else False,
                devmode=True if self.confinement == Confinement.DEVMODE else False,
                dangerous=self.dangerous,
            )

        if self.connections is not None:
            for connection in self.connections:
                connection.connect()

        if self.aliases is not None:
            for alias in self.aliases:
                alias.alias()

    def _dumps(self) -> Dict[str, str]:
        """Prepare Snap object for injection.

        Raises:
            FileNotFoundError: Raised if local snap package is not found.

        Returns:
            (Dict[str, str]):
                checksum (str): Checksum to verify authenticity of serialized Snap object.
                data (str): Base64 encoded string containing serialized Snap object.
                injectable (str): Injectable to run inside remote environment.
        """
        if self.local_snaps is not None:
            for local_snap in self.local_snaps:
                snap_path = pathlib.Path(local_snap)
                if not snap_path.exists() or not snap_path.is_file():
                    raise FileNotFoundError(
                        f"Could not find local snap package {snap_path}"
                    )
                self._cached_local_snaps.add(snap_path.read_bytes())

        return super()._dumps()

    def _injectable(self, data: Dict[str, str], **kwargs) -> str:
        """Generate injectable script that will be used to install snap packages.

        Args:
            data (Dict[str, str]): Data that needs to be in injectable script.
                - checksum (str): SHA224 checksum to verify authenticity of Snap object.
                - data (str): Base64 encoded Snap object to inject.

        Returns:
            (str): Injectable script.
        """
        return textwrap.dedent(
            f"""
            #!/usr/bin/env python3

            from {self.__module__} import {self.__class__.__name__}

            holder = {self.__class__.__name__}._loads("{data['checksum']}", "{data['data']}")
            holder._run()
            """
        ).strip("\n")



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

import pathlib

from setuptools import find_packages, setup

top_dir = pathlib.Path(__file__).parent
long_description = top_dir.joinpath("README.md").read_text()


setup(
    name="cleantest",
    version="0.4.0-rc1",
    description="Clean tests for developers in a hurry",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jason C. Nucciarone",
    author_email="jason.nucciarone@canonical.com",
    license="Apache-2.0",
    url="https://github.com/NucciTheBoss/cleantest",
    python_requires=">=3.6",
    packages=find_packages(
        where="src",
        include=["cleantest*"],
    ),
    package_dir={"": "src"},
    install_requires=[
        "pylxd",
    ],
    keywords=[
        "testing",
        "framework",
        "continuous integration",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Configure the flow of cleantest based on the test environment provider used."""

from .lxd._lxd_configurer import LXDConfigurer


class UnknownConfigurerError(Exception):
    """Raised when an unknown configurer option is passed to Configure."""


def Configure(configurer: str = "lxd") -> LXDConfigurer:  # noqa N802
    """Configure cleantest based on the test environment provider being used.

    Args:
        configurer (str): Configurer to use. Defaults to "lxd".

    Raises:
       UnknownConfigurerError: Raised if unknown configurer is specified.

    Returns:
        (LXDConfigurer): Configurer for LXD test environment provider.
    """
    dispatch = {"lxd": LXDConfigurer}
    if configurer not in dispatch.keys():
        raise UnknownConfigurerError(f"{configurer} is not a valid configurer option.")

    return dispatch[configurer]()



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Utility for interacting with systemd."""

import re
import subprocess
from shutil import which
from typing import Iterable, List, Optional, Union

from cleantest.meta import Result
from cleantest.meta.utils import detect_os_variant


class SystemdError(Exception):
    """Raise when an error is encountered with systemctl."""


def _check_systemd_available() -> None:
    """Check if systemd is available on test environment instance.

    Raises:
        SystemdError: Raised if systemctl executable is not found.
    """
    if which("systemctl") is None:
        raise SystemdError(f"systemd is not available on {detect_os_variant()}")


def _systemctl(
    command: str,
    services: Optional[Union[str, List[str]]] = None,
    optargs: Optional[List[str]] = None,
) -> Result:
    """Wrap systemctl commands for unit management.

    Args:
        command (str): systemctl command to execute.
        services (Union[str, List[str]]): Names of units to operate on.
        optargs (Optional[List[str]]): Optional arguments to pass to systemctl.

    Raises:
        SystemdError: Raised if executed systemctl command fails.

    Returns:
        (Result): Captured exit code, stdout, and stderr from subprocess.
    """
    _check_systemd_available()
    services = (
        [] if services is None else [services] if services == type(str) else services
    )
    optargs = optargs if optargs is not None else []
    _cmd = ["systemctl", command, *optargs, *services]
    try:
        process = subprocess.run(
            _cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        return Result(process.returncode, process.stdout, process.stderr)
    except subprocess.CalledProcessError:
        SystemdError(
            (
                f"Could not execute command {' '.join(_cmd)} on the following units: "
                ", ".join(services)
            )
        )


def is_active(*units: str) -> Iterable[bool]:
    """Check whether units are active.

    Args:
        *units (str): Units to operate on.

    Yields:
        (Iterable[bool]): Output of `systemctl is-active ...`.
            True if unit is active. False if unit is not active.
    """
    if len(units) == 0:
        raise SystemdError("No units to operate on.")
    for res in (
        _systemctl("is-active", [*units], ["--quiet"]).stdout.strip("\n").split("\n")
    ):
        yield True if res == "active" else False


def start(*units: str) -> None:
    """Start (activate) one or more units.

    Args:
        *units (str): Units to operate on.

    Raises:
        SystemdError: Raised if unit fails to start.
    """
    if len(units) == 0:
        raise SystemdError("No units to operate on.")
    for res in _systemctl("start", [*units]).stderr.strip("\n").split("\n"):
        if re.match(r"^Failed to start", res):
            raise SystemdError(res)


def stop(*units: str) -> None:
    """Stop (deactivate) one or more units.

    Args:
        *units (str): Units to operate on.

    Raises:
        SystemdError: Raised if unit fails to stop.
    """
    if len(units) == 0:
        raise SystemdError("No units to operate on.")
    for res in _systemctl("stop", [*units]).stderr.strip("\n").split("\n"):
        if re.match(r"^Failed to stop", res):
            raise SystemdError(res)


def kill(*units: str) -> None:
    """Send kill signal to units.

    Args:
        *units (str): Units to operate on.

    Raises:
        SystemdError: Raised if systemd fails to kill unit.
    """
    if len(units) == 0:
        raise SystemdError("No units to operate on.")
    for res in _systemctl("kill", [*units]).stderr.strip("\n").split("\n"):
        if re.match(r"^Failed to kill", res):
            raise SystemdError(res)


def clean(*units: str) -> None:
    """Clean runtime, cache, state, logs or configuration of units.

    Args:
        *units (str): Units to operate on.

    Raises:
        SystemdError: Raised if systemd fails to clean unit.
    """
    if len(units) == 0:
        raise SystemdError("No units to operate on.")
    for res in _systemctl("clean", [*units]).stderr.strip("\n").split("\n"):
        if re.match(r"^Failed to clean", res):
            raise SystemdError(res)


def restart(*units: str) -> None:
    """Start or restart units.

    Args:
        *units (str): Units to operate on.

    Raises:
        SystemdError: Raised if unit fails to start or restart.
    """
    if len(units) == 0:
        raise SystemdError("No units to operate on.")
    for res in _systemctl("restart", [*units]).stderr.strip("\n").split("\n"):
        if re.match(r"^Failed to restart", res):
            raise SystemdError(res)


def reload(*units: str, try_restart: bool = False) -> None:
    """Reload units.

    Args:
        *units (str): Units to operate on.
        try_restart (bool):
            Try `systemctl restart` if reload is not applicable for unit.

    Raises:
        SystemdError: Raised if unit fails to restart or reload.
    """
    if len(units) == 0:
        raise SystemdError("No units to operate on.")
    for unit in units:
        res = _systemctl("reload", unit).stderr.strip("\n")
        if re.match(r"^Failed to reload", res) and try_restart:
            restart(unit)
        else:
            raise SystemdError(res)


def pause(*units: str) -> None:
    """Pause units.

    Args:
        *units (str): Units to operate on.

    Raises:
        SystemdError: Raised if systemd fails to pause unit.
    """
    if len(units) == 0:
        raise SystemdError("No units to operate on.")
    _systemctl("disable", [*units], ["--now"])
    _systemctl("mask", [*units])
    for unit, status in zip([*units], is_active(*units)):
        if status:
            raise SystemdError(f"Failed to pause {unit}; it is still running.")


def resume(*units: str) -> None:
    """Resume units.

    Args:
        *units (str): Units to operate on.

    Raises:
        SystemdError: Raised if systemd fails to resume unit.
    """
    if len(units) == 0:
        raise SystemdError("No units to operate on.")
    _systemctl("unmask", [*units])
    _systemctl("enable", [*units], ["--now"])
    for unit, status in zip([*units], is_active(*units)):
        if not status:
            raise SystemdError(f"Failed to resume {unit}; it is not running.")


def daemon_reload() -> None:
    """Reload systemd configuration.

    Raises:
        SystemdError: Raised if systemd configuration fails to reload.
    """
    _systemctl("daemon-reload")



#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Test environment providers that can be used by cleantest."""

from .lxd.lxd import lxd
from .lxd.lxd_archon import LXDArchon
